import datetime
from collections import defaultdict
from collections.abc import Mapping, Sequence
from typing import TYPE_CHECKING, AbstractSet, Optional, Union, cast  # noqa: UP035

import dagster_shared.seven as seven
from dagster import (
    AssetKey,
    DagsterEventType,
    DagsterInstance,
    DagsterRun,
    MultiPartitionsDefinition,
    _check as check,
)
from dagster._core.definitions.data_time import CachingDataTimeResolver
from dagster._core.definitions.partition import (
    CachingDynamicPartitionsLoader,
    PartitionsDefinition,
    PartitionsSubset,
)
from dagster._core.definitions.remote_asset_graph import RemoteAssetNode
from dagster._core.definitions.time_window_partitions import (
    PartitionRangeStatus,
    TimeWindowPartitionsDefinition,
    TimeWindowPartitionsSubset,
    fetch_flattened_time_window_ranges,
)
from dagster._core.event_api import AssetRecordsFilter, EventLogRecord
from dagster._core.events.log import EventLogEntry
from dagster._core.instance import DynamicPartitionsStore
from dagster._core.loader import LoadingContext
from dagster._core.remote_representation.external import RemoteRepository
from dagster._core.storage.event_log.base import AssetRecord
from dagster._core.storage.event_log.sql_event_log import get_max_event_records_limit
from dagster._core.storage.partition_status_cache import (
    build_failed_and_in_progress_partition_subset,
    get_and_update_asset_status_cache_value,
    get_last_planned_storage_id,
    get_materialized_multipartitions,
    get_validated_partition_keys,
    is_cacheable_partition_type,
)
from dagster._time import get_current_datetime

from dagster_graphql.implementation.loader import StaleStatusLoader

if TYPE_CHECKING:
    from dagster_graphql.schema.asset_graph import (
        GrapheneAssetNode,
        GrapheneAssetNodeDefinitionCollision,
    )
    from dagster_graphql.schema.errors import GrapheneAssetNotFoundError
    from dagster_graphql.schema.freshness_policy import GrapheneAssetFreshnessInfo
    from dagster_graphql.schema.pipelines.pipeline import (
        GrapheneAsset,
        GrapheneDefaultPartitionStatuses,
        GrapheneMultiPartitionStatuses,
        GrapheneTimePartitionStatuses,
    )
    from dagster_graphql.schema.roots.assets import (
        GrapheneAssetConnection,
        GrapheneAssetRecordConnection,
    )
    from dagster_graphql.schema.util import ResolveInfo


def _normalize_asset_cursor_str(cursor_string: Optional[str]) -> Optional[str]:
    # the cursor for assets is derived from a json serialized string of the path.  Because there are
    # json serialization differences between JS and Python in its treatment of whitespace, we should
    # take extra precaution here and do a deserialization/serialization pass

    if not cursor_string:
        return cursor_string

    try:
        return seven.json.dumps(seven.json.loads(cursor_string))
    except seven.JSONDecodeError:
        return cursor_string


def get_asset_records(
    graphene_info: "ResolveInfo",
    prefix: Optional[Sequence[str]] = None,
    cursor: Optional[str] = None,
    limit: Optional[int] = None,
) -> "GrapheneAssetRecordConnection":
    from dagster_graphql.schema.pipelines.pipeline import GrapheneAssetRecord
    from dagster_graphql.schema.roots.assets import GrapheneAssetRecordConnection

    instance = graphene_info.context.instance

    normalized_cursor_str = _normalize_asset_cursor_str(cursor)
    materialized_assets = sorted(
        # TODO(salazarm): Replace this with `get_asset_records` once that supports pagination arguments.
        instance.get_asset_keys(prefix=prefix, limit=limit, cursor=normalized_cursor_str),
        key=str,
    )

    return GrapheneAssetRecordConnection(
        assets=[
            GrapheneAssetRecord(
                id=asset_key.to_string(),
                key=asset_key,
            )
            for asset_key in materialized_assets
        ],
        cursor=materialized_assets[-1].to_string() if materialized_assets else None,
    )


def get_assets(
    graphene_info: "ResolveInfo",
    prefix: Optional[Sequence[str]] = None,
    cursor: Optional[str] = None,
    limit: Optional[int] = None,
) -> "GrapheneAssetConnection":
    from dagster_graphql.schema.pipelines.pipeline import GrapheneAsset
    from dagster_graphql.schema.roots.assets import GrapheneAssetConnection

    instance = graphene_info.context.instance

    normalized_cursor_str = _normalize_asset_cursor_str(cursor)
    materialized_keys = instance.get_asset_keys(
        prefix=prefix, limit=limit, cursor=normalized_cursor_str
    )
    asset_nodes_by_asset_key = {
        asset_key: asset_node
        for asset_key, asset_node in get_asset_nodes_by_asset_key(graphene_info).items()
        if (not prefix or asset_key.path[: len(prefix)] == prefix)
        and (not normalized_cursor_str or asset_key.to_string() > normalized_cursor_str)
    }

    asset_keys = sorted(set(materialized_keys).union(asset_nodes_by_asset_key.keys()), key=str)
    if limit:
        asset_keys = asset_keys[:limit]

    return GrapheneAssetConnection(
        nodes=[
            GrapheneAsset(
                key=asset_key,
                definition=asset_nodes_by_asset_key.get(asset_key),
            )
            for asset_key in asset_keys
        ],
        cursor=asset_keys[-1].to_string() if asset_keys else None,
    )


def get_additional_required_keys(
    graphene_info: "ResolveInfo",
    asset_keys: AbstractSet[AssetKey],
) -> list["AssetKey"]:
    asset_nodes = {graphene_info.context.asset_graph.get(asset_key) for asset_key in asset_keys}

    required_asset_keys = set()
    for asset_node in asset_nodes:
        required_asset_keys.update(asset_node.execution_set_asset_keys)

    return list(required_asset_keys - asset_keys)


def get_asset_node_definition_collisions(
    graphene_info: "ResolveInfo", asset_keys: AbstractSet[AssetKey]
) -> list["GrapheneAssetNodeDefinitionCollision"]:
    from dagster_graphql.schema.asset_graph import GrapheneAssetNodeDefinitionCollision
    from dagster_graphql.schema.external import GrapheneRepository

    repos: dict[AssetKey, list[GrapheneRepository]] = defaultdict(list)
    for asset_key in asset_keys:
        remote_asset_node = graphene_info.context.asset_graph.get(asset_key)
        for info in remote_asset_node.repo_scoped_asset_infos:
            asset_node_snap = info.asset_node.asset_node_snap
            is_defined = (
                asset_node_snap.node_definition_name
                or asset_node_snap.graph_name
                or asset_node_snap.op_name
            )
            if not is_defined:
                continue

            repos[asset_node_snap.asset_key].append(GrapheneRepository(info.handle))

    results: list[GrapheneAssetNodeDefinitionCollision] = []
    for asset_key in repos.keys():
        if len(repos[asset_key]) > 1:
            results.append(
                GrapheneAssetNodeDefinitionCollision(
                    assetKey=asset_key, repositories=repos[asset_key]
                )
            )

    return results


def _graphene_asset_node(
    graphene_info: "ResolveInfo",
    remote_node: RemoteAssetNode,
    stale_status_loader: Optional[StaleStatusLoader],
    dynamic_partitions_loader: CachingDynamicPartitionsLoader,
):
    from dagster_graphql.schema.asset_graph import GrapheneAssetNode

    return GrapheneAssetNode(
        remote_node=remote_node,
        stale_status_loader=stale_status_loader,
        dynamic_partitions_loader=dynamic_partitions_loader,
    )


def get_asset_nodes_by_asset_key(
    graphene_info: "ResolveInfo",
) -> Mapping[AssetKey, "GrapheneAssetNode"]:
    """If multiple repositories have asset nodes for the same asset key, chooses the asset node that
    has an op.
    """
    stale_status_loader = StaleStatusLoader(
        instance=graphene_info.context.instance,
        asset_graph=lambda: graphene_info.context.asset_graph,
        loading_context=graphene_info.context,
    )

    dynamic_partitions_loader = CachingDynamicPartitionsLoader(graphene_info.context.instance)

    return {
        remote_node.key: _graphene_asset_node(
            graphene_info,
            remote_node,
            stale_status_loader=stale_status_loader,
            dynamic_partitions_loader=dynamic_partitions_loader,
        )
        for remote_node in graphene_info.context.asset_graph.asset_nodes
    }


def get_asset_node(
    graphene_info: "ResolveInfo", asset_key: AssetKey
) -> Union["GrapheneAssetNode", "GrapheneAssetNotFoundError"]:
    from dagster_graphql.schema.errors import GrapheneAssetNotFoundError

    check.inst_param(asset_key, "asset_key", AssetKey)

    if not graphene_info.context.asset_graph.has(asset_key):
        return GrapheneAssetNotFoundError(asset_key=asset_key)

    remote_node = graphene_info.context.asset_graph.get(asset_key)
    return _graphene_asset_node(
        graphene_info,
        remote_node,
        stale_status_loader=StaleStatusLoader(
            instance=graphene_info.context.instance,
            asset_graph=lambda: graphene_info.context.asset_graph,
            loading_context=graphene_info.context,
        ),
        dynamic_partitions_loader=CachingDynamicPartitionsLoader(
            graphene_info.context.instance,
        ),
    )


def get_asset(
    graphene_info: "ResolveInfo", asset_key: AssetKey
) -> Union["GrapheneAsset", "GrapheneAssetNotFoundError"]:
    from dagster_graphql.schema.errors import GrapheneAssetNotFoundError
    from dagster_graphql.schema.pipelines.pipeline import GrapheneAsset

    check.inst_param(asset_key, "asset_key", AssetKey)
    instance = graphene_info.context.instance

    has_remote_node = graphene_info.context.asset_graph.has(asset_key)
    if not has_remote_node and not instance.has_asset_key(asset_key):
        return GrapheneAssetNotFoundError(asset_key=asset_key)

    if has_remote_node:
        def_node = _graphene_asset_node(
            graphene_info,
            graphene_info.context.asset_graph.get(asset_key),
            stale_status_loader=None,
            dynamic_partitions_loader=CachingDynamicPartitionsLoader(
                graphene_info.context.instance,
            ),
        )
    else:
        def_node = None

    return GrapheneAsset(key=asset_key, definition=def_node)


def get_asset_materialization_event_records(
    graphene_info: "ResolveInfo",
    asset_key: AssetKey,
    partitions: Optional[Sequence[str]] = None,
    limit: Optional[int] = None,
    before_timestamp: Optional[float] = None,
    after_timestamp: Optional[float] = None,
    storage_ids: Optional[Sequence[int]] = None,
    cursor: Optional[str] = None,
) -> Sequence[EventLogRecord]:
    check.inst_param(asset_key, "asset_key", AssetKey)
    check.opt_int_param(limit, "limit")
    check.opt_float_param(before_timestamp, "before_timestamp")

    instance = graphene_info.context.instance
    records_filter = AssetRecordsFilter(
        asset_key=asset_key,
        asset_partitions=partitions,
        before_timestamp=before_timestamp,
        after_timestamp=after_timestamp,
        storage_ids=storage_ids,
    )
    if limit is None:
        event_records = []
        while True:
            event_records_result = instance.fetch_materializations(
                records_filter=records_filter,
                cursor=cursor,
                limit=get_max_event_records_limit(),
            )
            cursor = event_records_result.cursor
            event_records.extend(event_records_result.records)
            if not event_records_result.has_more:
                break
    else:
        event_records = instance.fetch_materializations(
            records_filter=records_filter, limit=limit, cursor=cursor
        ).records

    return event_records


def get_asset_materializations(
    graphene_info: "ResolveInfo",
    asset_key: AssetKey,
    partitions: Optional[Sequence[str]] = None,
    limit: Optional[int] = None,
    before_timestamp: Optional[float] = None,
    after_timestamp: Optional[float] = None,
    storage_ids: Optional[Sequence[int]] = None,
) -> Sequence[EventLogEntry]:
    return [
        event_record.event_log_entry
        for event_record in get_asset_materialization_event_records(
            graphene_info,
            asset_key,
            partitions,
            limit,
            before_timestamp,
            after_timestamp,
            storage_ids,
        )
    ]


def get_asset_failed_to_materialize_event_records(
    graphene_info: "ResolveInfo",
    asset_key: AssetKey,
    partitions: Optional[Sequence[str]] = None,
    limit: Optional[int] = None,
    before_timestamp: Optional[float] = None,
    after_timestamp: Optional[float] = None,
    storage_ids: Optional[Sequence[int]] = None,
    cursor: Optional[str] = None,
) -> Sequence[EventLogRecord]:
    check.inst_param(asset_key, "asset_key", AssetKey)
    check.opt_int_param(limit, "limit")
    check.opt_float_param(before_timestamp, "before_timestamp")

    instance = graphene_info.context.instance
    records_filter = AssetRecordsFilter(
        asset_key=asset_key,
        asset_partitions=partitions,
        before_timestamp=before_timestamp,
        after_timestamp=after_timestamp,
        storage_ids=storage_ids,
    )
    if limit is None:
        event_records = []
        while True:
            event_records_result = instance.fetch_failed_materializations(
                records_filter=records_filter, limit=get_max_event_records_limit(), cursor=cursor
            )
            cursor = event_records_result.cursor
            event_records.extend(event_records_result.records)
            if not event_records_result.has_more:
                break
    else:
        event_records = instance.fetch_failed_materializations(
            records_filter=records_filter, limit=limit, cursor=cursor
        ).records

    return event_records


def get_asset_observations(
    graphene_info: "ResolveInfo",
    asset_key: AssetKey,
    partitions: Optional[Sequence[str]] = None,
    limit: Optional[int] = None,
    before_timestamp: Optional[float] = None,
    after_timestamp: Optional[float] = None,
) -> Sequence[EventLogEntry]:
    check.inst_param(asset_key, "asset_key", AssetKey)
    check.opt_int_param(limit, "limit")
    check.opt_float_param(before_timestamp, "before_timestamp")
    check.opt_float_param(after_timestamp, "after_timestamp")

    instance = graphene_info.context.instance
    records_filter = AssetRecordsFilter(
        asset_key=asset_key,
        asset_partitions=partitions,
        before_timestamp=before_timestamp,
        after_timestamp=after_timestamp,
    )
    if limit is None:
        event_records = []
        cursor = None
        while True:
            event_records_result = instance.fetch_observations(
                records_filter=records_filter,
                cursor=cursor,
                limit=get_max_event_records_limit(),
            )
            cursor = event_records_result.cursor
            event_records.extend(event_records_result.records)
            if not event_records_result.has_more:
                break
    else:
        event_records = instance.fetch_observations(
            records_filter=records_filter, limit=limit
        ).records

    return [event_record.event_log_entry for event_record in event_records]


def get_assets_for_run(graphene_info: "ResolveInfo", run: DagsterRun) -> Sequence["GrapheneAsset"]:
    from dagster_graphql.schema.pipelines.pipeline import GrapheneAsset

    run_id = run.run_id

    # Fetch observations and materialization events from run
    event_log_records = [
        *graphene_info.context.instance.get_records_for_run(
            run_id,
            of_type=DagsterEventType.ASSET_MATERIALIZATION,
        ).records,
        *graphene_info.context.instance.get_records_for_run(
            run_id,
            of_type=DagsterEventType.ASSET_OBSERVATION,
        ).records,
    ]

    asset_keys = {
        record.event_log_entry.dagster_event.asset_key
        for record in event_log_records
        if record.event_log_entry.dagster_event
    }

    # Fetch planned asset keys from execution plan snapshot
    if run.execution_plan_snapshot_id:
        execution_plan_snapshot = check.not_none(
            graphene_info.context.instance.get_execution_plan_snapshot(
                run.execution_plan_snapshot_id
            )
        )
        asset_keys.update(execution_plan_snapshot.asset_selection)

    return [GrapheneAsset(key=asset_key) for asset_key in asset_keys if asset_key]


def get_unique_asset_id(
    asset_key: AssetKey,
    repository_location_name: Optional[str] = None,
    repository_name: Optional[str] = None,
) -> str:
    repository_identifier = (
        f"{repository_location_name}.{repository_name}"
        if repository_location_name and repository_name
        else ""
    )
    return (
        f"{repository_identifier}.{asset_key.to_string()}"
        if repository_identifier
        else f"{asset_key.to_string()}"
    )


def get_partition_subsets(
    instance: DagsterInstance,
    loading_context: LoadingContext,
    asset_key: AssetKey,
    dynamic_partitions_loader: DynamicPartitionsStore,
    partitions_def: Optional[PartitionsDefinition] = None,
) -> tuple[Optional[PartitionsSubset], Optional[PartitionsSubset], Optional[PartitionsSubset]]:
    """Returns a tuple of PartitionSubset objects: the first is the materialized partitions,
    the second is the failed partitions, and the third are in progress.
    """
    if not partitions_def:
        return None, None, None

    if instance.can_read_asset_status_cache() and is_cacheable_partition_type(partitions_def):
        # When the "cached_status_data" column exists in storage, update the column to contain
        # the latest partition status values
        updated_cache_value = get_and_update_asset_status_cache_value(
            instance,
            asset_key,
            partitions_def,
            dynamic_partitions_loader,
            loading_context,
        )
        materialized_subset = (
            updated_cache_value.deserialize_materialized_partition_subsets(partitions_def)
            if updated_cache_value
            else partitions_def.empty_subset()
        )
        failed_subset = (
            updated_cache_value.deserialize_failed_partition_subsets(partitions_def)
            if updated_cache_value
            else partitions_def.empty_subset()
        )
        in_progress_subset = (
            updated_cache_value.deserialize_in_progress_partition_subsets(partitions_def)
            if updated_cache_value
            else partitions_def.empty_subset()
        )

        return materialized_subset, failed_subset, in_progress_subset

    else:
        # If the partition status can't be cached, fetch partition status from storage
        if isinstance(partitions_def, MultiPartitionsDefinition):
            materialized_keys = get_materialized_multipartitions(
                instance, asset_key, partitions_def
            )
        else:
            materialized_keys = instance.get_materialized_partitions(asset_key)

        validated_keys = get_validated_partition_keys(
            dynamic_partitions_loader, partitions_def, set(materialized_keys)
        )

        materialized_subset = (
            partitions_def.empty_subset().with_partition_keys(validated_keys)
            if validated_keys
            else partitions_def.empty_subset()
        )

        asset_record = AssetRecord.blocking_get(loading_context, asset_key)

        failed_subset, in_progress_subset, _ = build_failed_and_in_progress_partition_subset(
            instance,
            asset_key,
            partitions_def,
            dynamic_partitions_loader,
            last_planned_materialization_storage_id=get_last_planned_storage_id(
                instance, asset_key, asset_record
            ),
        )

        return materialized_subset, failed_subset, in_progress_subset


def build_partition_statuses(
    dynamic_partitions_store: DynamicPartitionsStore,
    materialized_partitions_subset: Optional[PartitionsSubset],
    failed_partitions_subset: Optional[PartitionsSubset],
    in_progress_partitions_subset: Optional[PartitionsSubset],
    partitions_def: Optional[PartitionsDefinition],
) -> Union[
    "GrapheneTimePartitionStatuses",
    "GrapheneDefaultPartitionStatuses",
    "GrapheneMultiPartitionStatuses",
]:
    from dagster_graphql.schema.pipelines.pipeline import (
        GrapheneDefaultPartitionStatuses,
        GrapheneTimePartitionRangeStatus,
        GrapheneTimePartitionStatuses,
    )

    if (
        materialized_partitions_subset is None
        and failed_partitions_subset is None
        and in_progress_partitions_subset is None
    ):
        return GrapheneDefaultPartitionStatuses(
            materializedPartitions=[],
            failedPartitions=[],
            unmaterializedPartitions=[],
            materializingPartitions=[],
        )

    materialized_partitions_subset = check.not_none(materialized_partitions_subset)
    failed_partitions_subset = check.not_none(failed_partitions_subset)
    in_progress_partitions_subset = check.not_none(in_progress_partitions_subset)
    check.invariant(
        type(materialized_partitions_subset)
        == type(failed_partitions_subset)
        == type(in_progress_partitions_subset),
        "Expected materialized_partitions_subset, failed_partitions_subset, and"
        " in_progress_partitions_subset to be of the same type",
    )

    if isinstance(materialized_partitions_subset, TimeWindowPartitionsSubset):
        ranges = fetch_flattened_time_window_ranges(
            {
                PartitionRangeStatus.MATERIALIZED: materialized_partitions_subset,
                PartitionRangeStatus.FAILED: cast(
                    "TimeWindowPartitionsSubset", failed_partitions_subset
                ),
                PartitionRangeStatus.MATERIALIZING: cast(
                    "TimeWindowPartitionsSubset", in_progress_partitions_subset
                ),
            },
        )
        graphene_ranges = []
        for r in ranges:
            partition_key_range = cast(
                "TimeWindowPartitionsDefinition",
                materialized_partitions_subset.partitions_def,
            ).get_partition_key_range_for_time_window(r.time_window)
            graphene_ranges.append(
                GrapheneTimePartitionRangeStatus(
                    startTime=r.time_window.start.timestamp(),
                    endTime=r.time_window.end.timestamp(),
                    startKey=partition_key_range.start,
                    endKey=partition_key_range.end,
                    status=r.status,
                )
            )
        return GrapheneTimePartitionStatuses(ranges=graphene_ranges)
    elif isinstance(partitions_def, MultiPartitionsDefinition):
        return get_2d_run_length_encoded_partitions(
            dynamic_partitions_store,
            materialized_partitions_subset,
            failed_partitions_subset,
            in_progress_partitions_subset,
            partitions_def,
        )
    elif partitions_def:
        materialized_keys = materialized_partitions_subset.get_partition_keys()
        failed_keys = failed_partitions_subset.get_partition_keys()
        in_progress_keys = in_progress_partitions_subset.get_partition_keys()

        return GrapheneDefaultPartitionStatuses(
            materializedPartitions=set(materialized_keys)
            - set(failed_keys)
            - set(in_progress_keys),
            failedPartitions=failed_keys,
            unmaterializedPartitions=materialized_partitions_subset.get_partition_keys_not_in_subset(
                partitions_def=partitions_def,
                dynamic_partitions_store=dynamic_partitions_store,
            ),
            materializingPartitions=in_progress_keys,
        )
    else:
        check.failed("Should not reach this point")


def get_2d_run_length_encoded_partitions(
    dynamic_partitions_store: DynamicPartitionsStore,
    materialized_partitions_subset: PartitionsSubset,
    failed_partitions_subset: PartitionsSubset,
    in_progress_partitions_subset: PartitionsSubset,
    partitions_def: MultiPartitionsDefinition,
) -> "GrapheneMultiPartitionStatuses":
    from dagster_graphql.schema.pipelines.pipeline import (
        GrapheneMultiPartitionRangeStatuses,
        GrapheneMultiPartitionStatuses,
    )

    current_time = get_current_datetime()

    check.invariant(
        isinstance(partitions_def, MultiPartitionsDefinition),
        "Partitions definition should be multipartitioned",
    )

    primary_dim = partitions_def.primary_dimension
    secondary_dim = partitions_def.secondary_dimension

    dim2_materialized_partition_subset_by_dim1: dict[str, PartitionsSubset] = defaultdict(
        lambda: secondary_dim.partitions_def.empty_subset()
    )
    for partition_key in materialized_partitions_subset.get_partition_keys():
        multipartition_key = partitions_def.get_partition_key_from_str(partition_key)
        dim2_materialized_partition_subset_by_dim1[
            multipartition_key.keys_by_dimension[primary_dim.name]
        ] = dim2_materialized_partition_subset_by_dim1[
            multipartition_key.keys_by_dimension[primary_dim.name]
        ].with_partition_keys([multipartition_key.keys_by_dimension[secondary_dim.name]])

    dim2_failed_partition_subset_by_dim1: dict[str, PartitionsSubset] = defaultdict(
        lambda: secondary_dim.partitions_def.empty_subset()
    )
    for partition_key in failed_partitions_subset.get_partition_keys():
        multipartition_key = partitions_def.get_partition_key_from_str(partition_key)
        dim2_failed_partition_subset_by_dim1[
            multipartition_key.keys_by_dimension[primary_dim.name]
        ] = dim2_failed_partition_subset_by_dim1[
            multipartition_key.keys_by_dimension[primary_dim.name]
        ].with_partition_keys([multipartition_key.keys_by_dimension[secondary_dim.name]])

    dim2_in_progress_partition_subset_by_dim1: dict[str, PartitionsSubset] = defaultdict(
        lambda: secondary_dim.partitions_def.empty_subset()
    )
    for partition_key in in_progress_partitions_subset.get_partition_keys():
        multipartition_key = partitions_def.get_partition_key_from_str(partition_key)
        dim2_in_progress_partition_subset_by_dim1[
            multipartition_key.keys_by_dimension[primary_dim.name]
        ] = dim2_in_progress_partition_subset_by_dim1[
            multipartition_key.keys_by_dimension[primary_dim.name]
        ].with_partition_keys([multipartition_key.keys_by_dimension[secondary_dim.name]])

    materialized_2d_ranges = []

    dim1_keys = primary_dim.partitions_def.get_partition_keys(
        current_time=current_time, dynamic_partitions_store=dynamic_partitions_store
    )
    unevaluated_idx = 0
    range_start_idx = 0  # pointer to first dim1 partition with same dim2 materialization status

    if (
        len(dim1_keys) == 0
        or len(
            secondary_dim.partitions_def.get_partition_keys(
                current_time=current_time,
                dynamic_partitions_store=dynamic_partitions_store,
            )
        )
        == 0
    ):
        return GrapheneMultiPartitionStatuses(ranges=[], primaryDimensionName=primary_dim.name)

    while unevaluated_idx <= len(dim1_keys):
        if (
            unevaluated_idx == len(dim1_keys)
            or dim2_materialized_partition_subset_by_dim1[dim1_keys[unevaluated_idx]]
            != dim2_materialized_partition_subset_by_dim1[dim1_keys[range_start_idx]]
            or dim2_failed_partition_subset_by_dim1[dim1_keys[unevaluated_idx]]
            != dim2_failed_partition_subset_by_dim1[dim1_keys[range_start_idx]]
            or dim2_in_progress_partition_subset_by_dim1[dim1_keys[unevaluated_idx]]
            != dim2_in_progress_partition_subset_by_dim1[dim1_keys[range_start_idx]]
        ):
            # Add new multipartition range if we've reached the end of the dim1 keys or if the
            # second dimension subsets are different than for the previous dim1 key
            if (
                len(dim2_materialized_partition_subset_by_dim1[dim1_keys[range_start_idx]]) > 0
                or len(dim2_failed_partition_subset_by_dim1[dim1_keys[range_start_idx]]) > 0
                or len(dim2_in_progress_partition_subset_by_dim1[dim1_keys[range_start_idx]]) > 0
            ):
                # Do not add to materialized_2d_ranges if the dim2 partition subset is empty
                start_key = dim1_keys[range_start_idx]
                end_key = dim1_keys[unevaluated_idx - 1]

                primary_partitions_def = primary_dim.partitions_def
                if isinstance(primary_partitions_def, TimeWindowPartitionsDefinition):
                    time_windows = cast(
                        "TimeWindowPartitionsDefinition", primary_partitions_def
                    ).time_windows_for_partition_keys(
                        frozenset([start_key, end_key]),
                        current_time=current_time,
                        validate=False,  # we already know these keys are in the partition set
                    )
                    start_time = time_windows[0].start.timestamp()
                    end_time = time_windows[-1].end.timestamp()
                else:
                    start_time = None
                    end_time = None

                materialized_2d_ranges.append(
                    GrapheneMultiPartitionRangeStatuses(
                        primaryDimStartKey=start_key,
                        primaryDimEndKey=end_key,
                        primaryDimStartTime=start_time,
                        primaryDimEndTime=end_time,
                        secondaryDim=build_partition_statuses(
                            dynamic_partitions_store,
                            dim2_materialized_partition_subset_by_dim1[start_key],
                            dim2_failed_partition_subset_by_dim1[start_key],
                            dim2_in_progress_partition_subset_by_dim1[start_key],
                            secondary_dim.partitions_def,
                        ),
                    )
                )
            range_start_idx = unevaluated_idx
        unevaluated_idx += 1

    return GrapheneMultiPartitionStatuses(
        ranges=materialized_2d_ranges, primaryDimensionName=primary_dim.name
    )


def get_freshness_info(
    asset_key: AssetKey,
    data_time_resolver: CachingDataTimeResolver,
) -> "GrapheneAssetFreshnessInfo":
    from dagster_graphql.schema.freshness_policy import GrapheneAssetFreshnessInfo

    current_time = datetime.datetime.now(tz=datetime.timezone.utc)
    result = data_time_resolver.get_minutes_overdue(asset_key, evaluation_time=current_time)
    return GrapheneAssetFreshnessInfo(
        currentLagMinutes=result.lag_minutes if result else None,
        currentMinutesLate=result.overdue_minutes if result else None,
        latestMaterializationMinutesLate=None,
    )


def unique_repos(
    remote_repositories: Sequence[RemoteRepository],
) -> Sequence[RemoteRepository]:
    repos = []
    used = set()
    for remote_repository in remote_repositories:
        repo_id = (
            remote_repository.handle.location_name,
            remote_repository.name,
        )
        if repo_id not in used:
            used.add(repo_id)
            repos.append(remote_repository)

    return repos



import graphene
from dagster._core.storage.compute_log_manager import CapturedLogData

from dagster_graphql.schema.util import non_null_list


def from_captured_log_data(log_data: CapturedLogData):
    return GrapheneCapturedLogs(
        logKey=log_data.log_key,
        stdout=log_data.stdout.decode("utf-8") if log_data.stdout else None,
        stderr=log_data.stderr.decode("utf-8") if log_data.stderr else None,
        cursor=log_data.cursor,
    )


class GrapheneCapturedLogs(graphene.ObjectType):
    logKey = non_null_list(graphene.String)
    stdout = graphene.Field(graphene.String)
    stderr = graphene.Field(graphene.String)
    cursor = graphene.Field(graphene.String)

    class Meta:
        name = "CapturedLogs"


class GrapheneCapturedLogsMetadata(graphene.ObjectType):
    stdoutDownloadUrl = graphene.Field(graphene.String)
    stdoutLocation = graphene.Field(graphene.String)
    stderrDownloadUrl = graphene.Field(graphene.String)
    stderrLocation = graphene.Field(graphene.String)

    class Meta:
        name = "CapturedLogsMetadata"



from collections.abc import Mapping, Sequence
from typing import Any, Optional, cast

import dagster._check as check
import graphene
from dagster import AssetCheckKey
from dagster._core.definitions.events import AssetKey
from dagster._core.definitions.partition import CachingDynamicPartitionsLoader
from dagster._core.definitions.remote_asset_graph import RemoteAssetGraph
from dagster._core.definitions.selector import (
    InstigatorSelector,
    RepositorySelector,
    ResourceSelector,
    ScheduleSelector,
    SensorSelector,
)
from dagster._core.errors import DagsterInvariantViolationError
from dagster._core.execution.backfill import BulkActionStatus
from dagster._core.nux import get_has_seen_nux
from dagster._core.remote_representation.external import CompoundID
from dagster._core.scheduler.instigation import InstigatorStatus, InstigatorType
from dagster._core.storage.event_log.base import AssetRecord
from dagster._core.workspace.permissions import Permissions

from dagster_graphql.implementation.execution.backfill import get_asset_backfill_preview
from dagster_graphql.implementation.external import (
    fetch_location_entry,
    fetch_location_statuses,
    fetch_repositories,
    fetch_repository,
    fetch_workspace,
    get_remote_job_or_raise,
)
from dagster_graphql.implementation.fetch_asset_checks import fetch_asset_check_executions
from dagster_graphql.implementation.fetch_asset_condition_evaluations import (
    fetch_asset_condition_evaluation_record_for_partition,
    fetch_asset_condition_evaluation_records_for_asset_key,
    fetch_asset_condition_evaluation_records_for_evaluation_id,
    fetch_true_partitions_for_evaluation_node,
)
from dagster_graphql.implementation.fetch_assets import (
    get_additional_required_keys,
    get_asset,
    get_asset_node,
    get_asset_node_definition_collisions,
    get_asset_records,
    get_assets,
)
from dagster_graphql.implementation.fetch_auto_materialize_asset_evaluations import (
    fetch_auto_materialize_asset_evaluations,
    fetch_auto_materialize_asset_evaluations_for_evaluation_id,
)
from dagster_graphql.implementation.fetch_backfills import get_backfill, get_backfills
from dagster_graphql.implementation.fetch_env_vars import get_utilized_env_vars_or_error
from dagster_graphql.implementation.fetch_instigators import (
    get_instigation_states_by_repository_id,
    get_instigator_state_by_selector,
)
from dagster_graphql.implementation.fetch_logs import get_captured_log_metadata
from dagster_graphql.implementation.fetch_partition_sets import (
    get_partition_set,
    get_partition_sets_or_error,
)
from dagster_graphql.implementation.fetch_pipelines import (
    get_job_snapshot_or_error_from_job_selector,
    get_job_snapshot_or_error_from_snap_or_selector,
    get_job_snapshot_or_error_from_snapshot_id,
)
from dagster_graphql.implementation.fetch_resources import (
    get_resource_or_error,
    get_top_level_resources_or_error,
)
from dagster_graphql.implementation.fetch_runs import (
    gen_run_by_id,
    get_assets_latest_info,
    get_execution_plan,
    get_logs_for_run,
    get_run_group,
    get_run_tag_keys,
    get_run_tags,
    get_runs_feed_count,
    get_runs_feed_entries,
    validate_pipeline_config,
)
from dagster_graphql.implementation.fetch_schedules import (
    get_schedule_or_error,
    get_scheduler_or_error,
    get_schedules_or_error,
)
from dagster_graphql.implementation.fetch_sensors import get_sensor_or_error, get_sensors_or_error
from dagster_graphql.implementation.fetch_solids import get_graph_or_error
from dagster_graphql.implementation.fetch_ticks import get_instigation_ticks
from dagster_graphql.implementation.loader import StaleStatusLoader
from dagster_graphql.implementation.run_config_schema import resolve_run_config_schema_or_error
from dagster_graphql.implementation.utils import (
    UserFacingGraphQLError,
    capture_error,
    graph_selector_from_graphql,
    pipeline_selector_from_graphql,
)
from dagster_graphql.schema.asset_checks import GrapheneAssetCheckExecution
from dagster_graphql.schema.asset_condition_evaluations import (
    GrapheneAssetConditionEvaluation,
    GrapheneAssetConditionEvaluationRecordsOrError,
)
from dagster_graphql.schema.asset_graph import (
    GrapheneAssetLatestInfo,
    GrapheneAssetNode,
    GrapheneAssetNodeDefinitionCollision,
    GrapheneAssetNodeOrError,
)
from dagster_graphql.schema.auto_materialize_asset_evaluations import (
    GrapheneAutoMaterializeAssetEvaluationRecordsOrError,
)
from dagster_graphql.schema.backfill import (
    GrapheneAssetPartitions,
    GrapheneBulkActionStatus,
    GraphenePartitionBackfillOrError,
    GraphenePartitionBackfillsOrError,
)
from dagster_graphql.schema.entity_key import GrapheneAssetKey
from dagster_graphql.schema.env_vars import GrapheneEnvVarWithConsumersListOrError
from dagster_graphql.schema.external import (
    GrapheneRepositoriesOrError,
    GrapheneRepositoryConnection,
    GrapheneRepositoryOrError,
    GrapheneWorkspaceLocationEntryOrError,
    GrapheneWorkspaceLocationStatusEntriesOrError,
    GrapheneWorkspaceOrError,
)
from dagster_graphql.schema.inputs import (
    GrapheneAssetBackfillPreviewParams,
    GrapheneAssetCheckHandleInput,
    GrapheneAssetGroupSelector,
    GrapheneAssetKeyInput,
    GrapheneBulkActionsFilter,
    GrapheneGraphSelector,
    GrapheneInstigationSelector,
    GraphenePipelineSelector,
    GrapheneRepositorySelector,
    GrapheneResourceSelector,
    GrapheneRunsFilter,
    GrapheneScheduleSelector,
    GrapheneSensorSelector,
)
from dagster_graphql.schema.instance import GrapheneInstance
from dagster_graphql.schema.instigation import (
    GrapheneInstigationStateOrError,
    GrapheneInstigationStatesOrError,
    GrapheneInstigationStatus,
    GrapheneInstigationTick,
    GrapheneInstigationTickStatus,
)
from dagster_graphql.schema.logs.compute_logs import (
    GrapheneCapturedLogs,
    GrapheneCapturedLogsMetadata,
    from_captured_log_data,
)
from dagster_graphql.schema.partition_sets import (
    GraphenePartitionSetOrError,
    GraphenePartitionSetsOrError,
)
from dagster_graphql.schema.permissions import GraphenePermission
from dagster_graphql.schema.pipelines.config_result import GraphenePipelineConfigValidationResult
from dagster_graphql.schema.pipelines.pipeline import (
    GrapheneEventConnectionOrError,
    GraphenePipeline,
    GrapheneRunOrError,
)
from dagster_graphql.schema.pipelines.resource import (
    GrapheneResource,
    GrapheneResourceConnection,
    GrapheneResourcesOrError,
)
from dagster_graphql.schema.pipelines.snapshot import GraphenePipelineSnapshotOrError
from dagster_graphql.schema.resources import (
    GrapheneResourceDetailsListOrError,
    GrapheneResourceDetailsOrError,
)
from dagster_graphql.schema.roots.assets import (
    GrapheneAssetOrError,
    GrapheneAssetRecordsOrError,
    GrapheneAssetsOrError,
)
from dagster_graphql.schema.roots.execution_plan import GrapheneExecutionPlanOrError
from dagster_graphql.schema.roots.pipeline import GrapheneGraphOrError, GraphenePipelineOrError
from dagster_graphql.schema.run_config import GrapheneRunConfigSchemaOrError
from dagster_graphql.schema.runs import (
    GrapheneRunConfigData,
    GrapheneRunGroupOrError,
    GrapheneRunIds,
    GrapheneRunIdsOrError,
    GrapheneRuns,
    GrapheneRunsOrError,
    GrapheneRunTagKeysOrError,
    GrapheneRunTagsOrError,
    parse_run_config_input,
)
from dagster_graphql.schema.runs_feed import (
    GrapheneRunsFeedConnectionOrError,
    GrapheneRunsFeedCount,
    GrapheneRunsFeedCountOrError,
    GrapheneRunsFeedView,
)
from dagster_graphql.schema.schedules import (
    GrapheneScheduleOrError,
    GrapheneSchedulerOrError,
    GrapheneSchedulesOrError,
)
from dagster_graphql.schema.sensors import GrapheneSensorOrError, GrapheneSensorsOrError
from dagster_graphql.schema.test import GrapheneTestFields
from dagster_graphql.schema.util import ResolveInfo, get_compute_log_manager, non_null_list


class GrapheneQuery(graphene.ObjectType):
    """The root for all queries to retrieve data from the Dagster instance."""

    class Meta:
        name = "Query"

    version = graphene.Field(
        graphene.NonNull(graphene.String),
        description="Retrieve the version of Dagster running in the Dagster deployment.",
    )

    repositoriesOrError = graphene.Field(
        graphene.NonNull(GrapheneRepositoriesOrError),
        repositorySelector=graphene.Argument(GrapheneRepositorySelector),
        description="Retrieve all the repositories.",
    )

    repositoryOrError = graphene.Field(
        graphene.NonNull(GrapheneRepositoryOrError),
        repositorySelector=graphene.NonNull(GrapheneRepositorySelector),
        description="Retrieve a repository by its location name and repository name.",
    )

    workspaceOrError = graphene.Field(
        graphene.NonNull(GrapheneWorkspaceOrError),
        description="Retrieve the workspace and its locations.",
    )

    locationStatusesOrError = graphene.Field(
        graphene.NonNull(GrapheneWorkspaceLocationStatusEntriesOrError),
        description="Retrieve location status for workspace locations.",
    )

    workspaceLocationEntryOrError = graphene.Field(
        GrapheneWorkspaceLocationEntryOrError,
        name=graphene.Argument(graphene.NonNull(graphene.String)),
        description="Retrieve a workspace entry by name.",
    )

    pipelineOrError = graphene.Field(
        graphene.NonNull(GraphenePipelineOrError),
        params=graphene.NonNull(GraphenePipelineSelector),
        description="Retrieve a job by its location name, repository name, and job name.",
    )

    resourcesOrError = graphene.Field(
        graphene.NonNull(GrapheneResourcesOrError),
        pipelineSelector=graphene.NonNull(GraphenePipelineSelector),
        description="Retrieve the list of resources for a given job.",
    )

    pipelineSnapshotOrError = graphene.Field(
        graphene.NonNull(GraphenePipelineSnapshotOrError),
        snapshotId=graphene.String(),
        activePipelineSelector=graphene.Argument(GraphenePipelineSelector),
        description=(
            "Retrieve a job snapshot by its id or location name, repository name, and job name."
        ),
    )

    graphOrError = graphene.Field(
        graphene.NonNull(GrapheneGraphOrError),
        selector=graphene.Argument(GrapheneGraphSelector),
        description="Retrieve a graph by its location name, repository name, and graph name.",
    )

    scheduler = graphene.Field(
        graphene.NonNull(GrapheneSchedulerOrError),
        description="Retrieve the name of the scheduler running in the Dagster deployment.",
    )

    scheduleOrError = graphene.Field(
        graphene.NonNull(GrapheneScheduleOrError),
        schedule_selector=graphene.NonNull(GrapheneScheduleSelector),
        description="Retrieve a schedule by its location name, repository name, and schedule name.",
    )

    schedulesOrError = graphene.Field(
        graphene.NonNull(GrapheneSchedulesOrError),
        repositorySelector=graphene.NonNull(GrapheneRepositorySelector),
        scheduleStatus=graphene.Argument(GrapheneInstigationStatus),
        description="Retrieve all the schedules.",
    )

    topLevelResourceDetailsOrError = graphene.Field(
        graphene.NonNull(GrapheneResourceDetailsOrError),
        resourceSelector=graphene.NonNull(GrapheneResourceSelector),
        description=(
            "Retrieve a top level resource by its location name, repository name, and resource"
            " name."
        ),
    )

    allTopLevelResourceDetailsOrError = graphene.Field(
        graphene.NonNull(GrapheneResourceDetailsListOrError),
        repositorySelector=graphene.NonNull(GrapheneRepositorySelector),
        description="Retrieve all the top level resources.",
    )

    utilizedEnvVarsOrError = graphene.Field(
        graphene.NonNull(GrapheneEnvVarWithConsumersListOrError),
        repositorySelector=graphene.NonNull(GrapheneRepositorySelector),
        description="Retrieve all the utilized environment variables for the given repo.",
    )

    sensorOrError = graphene.Field(
        graphene.NonNull(GrapheneSensorOrError),
        sensorSelector=graphene.NonNull(GrapheneSensorSelector),
        description="Retrieve a sensor by its location name, repository name, and sensor name.",
    )
    sensorsOrError = graphene.Field(
        graphene.NonNull(GrapheneSensorsOrError),
        repositorySelector=graphene.NonNull(GrapheneRepositorySelector),
        sensorStatus=graphene.Argument(GrapheneInstigationStatus),
        description="Retrieve all the sensors.",
    )

    instigationStateOrError = graphene.Field(
        graphene.NonNull(GrapheneInstigationStateOrError),
        instigationSelector=graphene.NonNull(GrapheneInstigationSelector),
        id=graphene.Argument(graphene.String),
        description=(
            "Retrieve the state for a schedule or sensor by its location name, repository name, and"
            " schedule/sensor name."
        ),
    )

    instigationStatesOrError = graphene.Field(
        graphene.NonNull(GrapheneInstigationStatesOrError),
        repositoryID=graphene.NonNull(graphene.String),
        description=(
            "Retrieve the state for a group of instigators (schedule/sensor) by their containing repository id."
        ),
    )

    partitionSetsOrError = graphene.Field(
        graphene.NonNull(GraphenePartitionSetsOrError),
        repositorySelector=graphene.NonNull(GrapheneRepositorySelector),
        pipelineName=graphene.NonNull(graphene.String),
        description=(
            "Retrieve the partition sets for a job by its location name, repository name, and job"
            " name."
        ),
    )
    partitionSetOrError = graphene.Field(
        graphene.NonNull(GraphenePartitionSetOrError),
        repositorySelector=graphene.NonNull(GrapheneRepositorySelector),
        partitionSetName=graphene.String(),
        description=(
            "Retrieve a partition set by its location name, repository name, and partition set"
            " name."
        ),
    )

    pipelineRunsOrError = graphene.Field(
        graphene.NonNull(GrapheneRunsOrError),
        filter=graphene.Argument(GrapheneRunsFilter),
        cursor=graphene.String(),
        limit=graphene.Int(),
        description="Retrieve runs after applying a filter, cursor, and limit.",
    )
    pipelineRunOrError = graphene.Field(
        graphene.NonNull(GrapheneRunOrError),
        runId=graphene.NonNull(graphene.ID),
        description="Retrieve a run by its run id.",
    )
    runsOrError = graphene.Field(
        graphene.NonNull(GrapheneRunsOrError),
        filter=graphene.Argument(GrapheneRunsFilter),
        cursor=graphene.String(),
        limit=graphene.Int(),
        description="Retrieve runs after applying a filter, cursor, and limit.",
    )
    runOrError = graphene.Field(
        graphene.NonNull(GrapheneRunOrError),
        runId=graphene.NonNull(graphene.ID),
        description="Retrieve a run by its run id.",
    )
    runsFeedOrError = graphene.Field(
        graphene.NonNull(GrapheneRunsFeedConnectionOrError),
        limit=graphene.NonNull(graphene.Int),
        cursor=graphene.String(),
        view=graphene.NonNull(GrapheneRunsFeedView),
        filter=graphene.Argument(GrapheneRunsFilter),
        description="Retrieve entries for the Runs Feed after applying a filter, cursor and limit.",
    )
    runsFeedCountOrError = graphene.Field(
        graphene.NonNull(GrapheneRunsFeedCountOrError),
        view=graphene.NonNull(GrapheneRunsFeedView),
        filter=graphene.Argument(GrapheneRunsFilter),
        description="Retrieve the number of entries for the Runs Feed after applying a filter.",
    )
    runTagKeysOrError = graphene.Field(
        GrapheneRunTagKeysOrError,
        description="Retrieve the distinct tag keys from all runs.",
    )
    runTagsOrError = graphene.Field(
        GrapheneRunTagsOrError,
        tagKeys=graphene.Argument(graphene.List(graphene.NonNull(graphene.String))),
        valuePrefix=graphene.String(),
        limit=graphene.Int(),
        description="Retrieve all the distinct key-value tags from all runs.",
    )
    runIdsOrError = graphene.Field(
        graphene.NonNull(GrapheneRunIdsOrError),
        filter=graphene.Argument(GrapheneRunsFilter),
        cursor=graphene.String(),
        limit=graphene.Int(),
        description="Retrieve run IDs after applying a filter, cursor, and limit.",
    )

    runGroupOrError = graphene.Field(
        graphene.NonNull(GrapheneRunGroupOrError),
        runId=graphene.NonNull(graphene.ID),
        description="Retrieve a group of runs with the matching root run id.",
    )

    isPipelineConfigValid = graphene.Field(
        graphene.NonNull(GraphenePipelineConfigValidationResult),
        pipeline=graphene.Argument(graphene.NonNull(GraphenePipelineSelector)),
        mode=graphene.Argument(graphene.NonNull(graphene.String)),
        runConfigData=graphene.Argument(GrapheneRunConfigData),
        description="Retrieve whether the run configuration is valid or invalid.",
    )

    executionPlanOrError = graphene.Field(
        graphene.NonNull(GrapheneExecutionPlanOrError),
        pipeline=graphene.Argument(graphene.NonNull(GraphenePipelineSelector)),
        mode=graphene.Argument(graphene.NonNull(graphene.String)),
        runConfigData=graphene.Argument(GrapheneRunConfigData),
        description="Retrieve the execution plan for a job and its run configuration.",
    )

    runConfigSchemaOrError = graphene.Field(
        graphene.NonNull(GrapheneRunConfigSchemaOrError),
        args={
            "selector": graphene.Argument(graphene.NonNull(GraphenePipelineSelector)),
            "mode": graphene.Argument(graphene.String),
        },
        description="Retrieve the run configuration schema for a job.",
    )

    instance = graphene.Field(
        graphene.NonNull(GrapheneInstance),
        description="Retrieve the instance configuration for the Dagster deployment.",
    )

    assetsOrError = graphene.Field(
        graphene.NonNull(GrapheneAssetsOrError),
        prefix=graphene.List(graphene.NonNull(graphene.String)),
        cursor=graphene.String(),
        limit=graphene.Int(),
        description="Retrieve all assets (both with or without a definition) after applying a prefix filter, cursor, and limit.",
    )

    assetRecordsOrError = graphene.Field(
        graphene.NonNull(GrapheneAssetRecordsOrError),
        prefix=graphene.List(graphene.NonNull(graphene.String)),
        cursor=graphene.String(),
        limit=graphene.Int(),
        description="Retrieve materialized asset records after applying a prefix filter, cursor, and limit.",
    )
    assetOrError = graphene.Field(
        graphene.NonNull(GrapheneAssetOrError),
        assetKey=graphene.Argument(graphene.NonNull(GrapheneAssetKeyInput)),
        description="Retrieve an asset by asset key.",
    )

    assetNodes = graphene.Field(
        non_null_list(GrapheneAssetNode),
        group=graphene.Argument(GrapheneAssetGroupSelector),
        pipeline=graphene.Argument(GraphenePipelineSelector),
        assetKeys=graphene.Argument(graphene.List(graphene.NonNull(GrapheneAssetKeyInput))),
        loadMaterializations=graphene.Boolean(default_value=False),
        description=(
            "Retrieve asset nodes (assets with a definition) after applying a filter on asset group, job, and asset keys."
        ),
    )

    assetNodeOrError = graphene.Field(
        graphene.NonNull(GrapheneAssetNodeOrError),
        assetKey=graphene.Argument(graphene.NonNull(GrapheneAssetKeyInput)),
        description="Retrieve an asset node by asset key.",
    )

    assetNodeAdditionalRequiredKeys = graphene.Field(
        non_null_list(GrapheneAssetKey),
        assetKeys=graphene.Argument(non_null_list(GrapheneAssetKeyInput)),
        description="Retrieve a list of additional asset keys that must be materialized with the provided selection (due to @multi_assets with can_subset=False constraints.)",
    )

    assetNodeDefinitionCollisions = graphene.Field(
        non_null_list(GrapheneAssetNodeDefinitionCollision),
        assetKeys=graphene.Argument(non_null_list(GrapheneAssetKeyInput)),
        description=(
            "Retrieve a list of asset keys where two or more repos provide an asset definition."
            " Note: Assets should "
        )
        + "not be defined in more than one repository - this query is used to present warnings and"
        " errors in the Dagster UI.",
    )

    partitionBackfillOrError = graphene.Field(
        graphene.NonNull(GraphenePartitionBackfillOrError),
        backfillId=graphene.Argument(graphene.NonNull(graphene.String)),
        description="Retrieve a backfill by backfill id.",
    )

    assetBackfillPreview = graphene.Field(
        non_null_list(GrapheneAssetPartitions),
        params=graphene.Argument(graphene.NonNull(GrapheneAssetBackfillPreviewParams)),
        description="Fetch the partitions that would be targeted by a backfill, given the root partitions targeted.",
    )

    partitionBackfillsOrError = graphene.Field(
        graphene.NonNull(GraphenePartitionBackfillsOrError),
        status=graphene.Argument(GrapheneBulkActionStatus),
        cursor=graphene.String(),
        limit=graphene.Int(),
        filters=graphene.Argument(GrapheneBulkActionsFilter),
        description="Retrieve backfills after applying a status filter, cursor, and limit.",
    )

    permissions = graphene.Field(
        non_null_list(GraphenePermission),
        description="Retrieve the set of permissions for the Dagster deployment.",
    )

    canBulkTerminate = graphene.Field(
        graphene.NonNull(graphene.Boolean),
        description="Returns whether the user has permission to terminate runs in the deployment",
    )

    assetsLatestInfo = graphene.Field(
        non_null_list(GrapheneAssetLatestInfo),
        assetKeys=graphene.Argument(non_null_list(GrapheneAssetKeyInput)),
        description="Retrieve the latest materializations for a set of assets by asset keys.",
    )

    logsForRun = graphene.Field(
        graphene.NonNull(GrapheneEventConnectionOrError),
        runId=graphene.NonNull(graphene.ID),
        afterCursor=graphene.String(),
        limit=graphene.Int(),
        description="Retrieve event logs after applying a run id filter, cursor, and limit.",
    )

    capturedLogsMetadata = graphene.Field(
        graphene.NonNull(GrapheneCapturedLogsMetadata),
        logKey=graphene.Argument(non_null_list(graphene.String)),
        description="Retrieve the captured log metadata for a given log key.",
    )
    capturedLogs = graphene.Field(
        graphene.NonNull(GrapheneCapturedLogs),
        logKey=graphene.Argument(non_null_list(graphene.String)),
        cursor=graphene.Argument(graphene.String),
        limit=graphene.Argument(graphene.Int),
        description="Captured logs are the stdout/stderr logs for a given log key",
    )

    shouldShowNux = graphene.Field(
        graphene.NonNull(graphene.Boolean),
        description="Whether or not the NUX should be shown to the user",
    )

    test = graphene.Field(
        GrapheneTestFields,
        description="Provides fields for testing behavior",
    )

    autoMaterializeAssetEvaluationsOrError = graphene.Field(
        GrapheneAutoMaterializeAssetEvaluationRecordsOrError,
        assetKey=graphene.Argument(graphene.NonNull(GrapheneAssetKeyInput)),
        limit=graphene.Argument(graphene.NonNull(graphene.Int)),
        cursor=graphene.Argument(graphene.String),
        description="Retrieve the auto materialization evaluation records for an asset.",
    )

    truePartitionsForAutomationConditionEvaluationNode = graphene.Field(
        non_null_list(graphene.String),
        assetKey=graphene.Argument(GrapheneAssetKeyInput),
        evaluationId=graphene.Argument(graphene.NonNull(graphene.ID)),
        nodeUniqueId=graphene.Argument(graphene.String),
        description="Retrieve the partition keys which were true for a specific automation condition evaluation node.",
    )

    autoMaterializeEvaluationsForEvaluationId = graphene.Field(
        GrapheneAutoMaterializeAssetEvaluationRecordsOrError,
        evaluationId=graphene.Argument(graphene.NonNull(graphene.ID)),
        description=(
            "Retrieve the auto materialization evaluation records for a given evaluation ID."
        ),
    )

    assetConditionEvaluationForPartition = graphene.Field(
        GrapheneAssetConditionEvaluation,
        assetKey=graphene.Argument(GrapheneAssetKeyInput),
        evaluationId=graphene.Argument(graphene.NonNull(graphene.ID)),
        partition=graphene.Argument(graphene.NonNull(graphene.String)),
        description="Retrieve the condition evaluation for an asset and partition.",
    )

    assetConditionEvaluationRecordsOrError = graphene.Field(
        GrapheneAssetConditionEvaluationRecordsOrError,
        assetKey=graphene.Argument(GrapheneAssetKeyInput),
        assetCheckKey=graphene.Argument(GrapheneAssetCheckHandleInput, required=False),
        limit=graphene.Argument(graphene.NonNull(graphene.Int)),
        cursor=graphene.Argument(graphene.String),
        description="Retrieve the condition evaluation records for an asset.",
    )

    assetConditionEvaluationsForEvaluationId = graphene.Field(
        GrapheneAssetConditionEvaluationRecordsOrError,
        evaluationId=graphene.Argument(graphene.NonNull(graphene.ID)),
        description=("Retrieve the condition evaluation records for a given evaluation ID."),
    )

    autoMaterializeTicks = graphene.Field(
        non_null_list(GrapheneInstigationTick),
        dayRange=graphene.Int(),
        dayOffset=graphene.Int(),
        limit=graphene.Int(),
        cursor=graphene.String(),
        statuses=graphene.List(graphene.NonNull(GrapheneInstigationTickStatus)),
        beforeTimestamp=graphene.Float(),
        afterTimestamp=graphene.Float(),
        description="Fetch the history of auto-materialization ticks",
    )

    assetCheckExecutions = graphene.Field(
        non_null_list(GrapheneAssetCheckExecution),
        assetKey=graphene.Argument(graphene.NonNull(GrapheneAssetKeyInput)),
        checkName=graphene.Argument(graphene.NonNull(graphene.String)),
        limit=graphene.NonNull(graphene.Int),
        cursor=graphene.String(),
        description="Retrieve the executions for a given asset check.",
    )

    @capture_error
    def resolve_repositoriesOrError(
        self,
        graphene_info: ResolveInfo,
        repositorySelector: Optional[GrapheneRepositorySelector] = None,
    ):
        if repositorySelector:
            return GrapheneRepositoryConnection(
                nodes=[
                    fetch_repository(
                        graphene_info,
                        RepositorySelector.from_graphql_input(repositorySelector),
                    )
                ]
            )
        return fetch_repositories(graphene_info)

    @capture_error
    def resolve_repositoryOrError(
        self, graphene_info: ResolveInfo, repositorySelector: GrapheneRepositorySelector
    ):
        return fetch_repository(
            graphene_info, RepositorySelector.from_graphql_input(repositorySelector)
        )

    @capture_error
    def resolve_workspaceOrError(self, graphene_info: ResolveInfo):
        return fetch_workspace(graphene_info.context)

    @capture_error
    def resolve_workspaceLocationEntryOrError(self, graphene_info: ResolveInfo, name: str):
        return fetch_location_entry(graphene_info.context, name)

    @capture_error
    def resolve_locationStatusesOrError(self, graphene_info: ResolveInfo):
        return fetch_location_statuses(graphene_info.context)

    @capture_error
    def resolve_pipelineSnapshotOrError(
        self,
        graphene_info: ResolveInfo,
        snapshotId: Optional[str] = None,
        activePipelineSelector: Optional[GraphenePipelineSelector] = None,
    ):
        if activePipelineSelector:
            job_selector = pipeline_selector_from_graphql(activePipelineSelector)
            if snapshotId:
                return get_job_snapshot_or_error_from_snap_or_selector(
                    graphene_info, job_selector, snapshotId
                )
            return get_job_snapshot_or_error_from_job_selector(graphene_info, job_selector)
        elif snapshotId:
            return get_job_snapshot_or_error_from_snapshot_id(graphene_info, snapshotId)
        else:
            raise DagsterInvariantViolationError(
                "Must pass snapshotId or activePipelineSelector",
            )

    @capture_error
    def resolve_graphOrError(
        self,
        graphene_info: ResolveInfo,
        selector: Optional[GrapheneGraphSelector] = None,
    ):
        if selector is None:
            raise DagsterInvariantViolationError(
                "Must pass graph selector",
            )

        graph_selector = graph_selector_from_graphql(selector)
        return get_graph_or_error(graphene_info, graph_selector)

    def resolve_version(self, graphene_info: ResolveInfo):
        return graphene_info.context.version

    @capture_error
    def resolve_scheduler(self, graphene_info: ResolveInfo):
        return get_scheduler_or_error(graphene_info)

    @capture_error
    def resolve_scheduleOrError(
        self, graphene_info: ResolveInfo, schedule_selector: GrapheneScheduleSelector
    ):
        return get_schedule_or_error(
            graphene_info, ScheduleSelector.from_graphql_input(schedule_selector)
        )

    @capture_error
    def resolve_schedulesOrError(
        self,
        graphene_info: ResolveInfo,
        repositorySelector: GrapheneRepositorySelector,
        scheduleStatus: Optional[GrapheneInstigationStatus] = None,
    ):
        if scheduleStatus == GrapheneInstigationStatus.RUNNING:
            instigator_statuses = {
                InstigatorStatus.RUNNING,
                InstigatorStatus.DECLARED_IN_CODE,
            }
        elif scheduleStatus == GrapheneInstigationStatus.STOPPED:
            instigator_statuses = {InstigatorStatus.STOPPED}
        else:
            instigator_statuses = None

        return get_schedules_or_error(
            graphene_info,
            RepositorySelector.from_graphql_input(repositorySelector),
            instigator_statuses,
        )

    @capture_error
    def resolve_topLevelResourceDetailsOrError(self, graphene_info: ResolveInfo, resourceSelector):
        return get_resource_or_error(
            graphene_info, ResourceSelector.from_graphql_input(resourceSelector)
        )

    def resolve_allTopLevelResourceDetailsOrError(self, graphene_info: ResolveInfo, **kwargs):
        return get_top_level_resources_or_error(
            graphene_info,
            RepositorySelector.from_graphql_input(kwargs.get("repositorySelector")),
        )

    @capture_error
    def resolve_utilizedEnvVarsOrError(self, graphene_info: ResolveInfo, **kwargs):
        return get_utilized_env_vars_or_error(
            graphene_info,
            RepositorySelector.from_graphql_input(kwargs.get("repositorySelector")),
        )

    @capture_error
    def resolve_sensorOrError(
        self, graphene_info: ResolveInfo, sensorSelector: GrapheneRepositorySelector
    ):
        return get_sensor_or_error(graphene_info, SensorSelector.from_graphql_input(sensorSelector))

    @capture_error
    def resolve_sensorsOrError(
        self,
        graphene_info,
        repositorySelector: GrapheneRepositorySelector,
        sensorStatus: Optional[GrapheneInstigationStatus] = None,
    ):
        if sensorStatus == GrapheneInstigationStatus.RUNNING:
            instigator_statuses = {
                InstigatorStatus.RUNNING,
                InstigatorStatus.DECLARED_IN_CODE,
            }
        elif sensorStatus == GrapheneInstigationStatus.STOPPED:
            instigator_statuses = {InstigatorStatus.STOPPED}
        else:
            instigator_statuses = None
        return get_sensors_or_error(
            graphene_info,
            RepositorySelector.from_graphql_input(repositorySelector),
            instigator_statuses,
        )

    @capture_error
    def resolve_instigationStateOrError(
        self,
        graphene_info: ResolveInfo,
        *,
        instigationSelector: GrapheneInstigationSelector,
        id: Optional[str] = None,
    ):
        return get_instigator_state_by_selector(
            graphene_info,
            InstigatorSelector.from_graphql_input(instigationSelector),
            CompoundID.from_string(id) if id else None,
        )

    @capture_error
    def resolve_instigationStatesOrError(
        self,
        graphene_info: ResolveInfo,
        repositoryID: str,
    ):
        return get_instigation_states_by_repository_id(
            graphene_info,
            CompoundID.from_string(repositoryID),
        )

    @capture_error
    def resolve_pipelineOrError(self, graphene_info: ResolveInfo, params: GraphenePipelineSelector):
        return GraphenePipeline(
            get_remote_job_or_raise(graphene_info, pipeline_selector_from_graphql(params))
        )

    @capture_error
    def resolve_resourcesOrError(
        self, graphene_info: ResolveInfo, pipelineSelector: GraphenePipelineSelector
    ) -> Sequence[GrapheneResource]:
        from dagster_graphql.schema.errors import GraphenePipelineNotFoundError

        job_selector = pipeline_selector_from_graphql(pipelineSelector)

        if not graphene_info.context.has_job(job_selector):
            raise UserFacingGraphQLError(GraphenePipelineNotFoundError(selector=job_selector))

        check.invariant(
            not job_selector.is_subset_selection,
            "resourcesOrError only accepts non-subsetted selectors",
        )

        def _get_config_type(key: str):
            return graphene_info.context.get_config_type(job_selector, key)

        return GrapheneResourceConnection(
            resources=[
                GrapheneResource(_get_config_type, resource_snap)
                for resource_snap in graphene_info.context.get_resources(job_selector)
            ]
        )

    def resolve_pipelineRunsOrError(
        self,
        _graphene_info: ResolveInfo,
        filter: Optional[GrapheneRunsFilter] = None,  # noqa: A002
        cursor: Optional[str] = None,
        limit: Optional[int] = None,
    ):
        selector = filter.to_selector() if filter is not None else None

        return GrapheneRuns(
            filters=selector,
            cursor=cursor,
            limit=limit,
        )

    async def resolve_pipelineRunOrError(self, graphene_info: ResolveInfo, runId: graphene.ID):
        return await gen_run_by_id(graphene_info, runId)

    def resolve_runsOrError(
        self,
        _graphene_info: ResolveInfo,
        filter: Optional[GrapheneRunsFilter] = None,  # noqa: A002
        cursor: Optional[str] = None,
        limit: Optional[int] = None,
    ):
        selector = filter.to_selector() if filter is not None else None

        return GrapheneRuns(
            filters=selector,
            cursor=cursor,
            limit=limit,
        )

    def resolve_runIdsOrError(
        self,
        _graphene_info: ResolveInfo,
        filter: Optional[GrapheneRunsFilter] = None,  # noqa: A002
        cursor: Optional[str] = None,
        limit: Optional[int] = None,
    ):
        selector = filter.to_selector() if filter is not None else None

        return GrapheneRunIds(
            filters=selector,
            cursor=cursor,
            limit=limit,
        )

    async def resolve_runOrError(self, graphene_info: ResolveInfo, runId):
        return await gen_run_by_id(graphene_info, runId)

    def resolve_runsFeedOrError(
        self,
        graphene_info: ResolveInfo,
        limit: int,
        view: GrapheneRunsFeedView,
        cursor: Optional[str] = None,
        filter: Optional[GrapheneRunsFilter] = None,  # noqa: A002
    ):
        selector = filter.to_selector() if filter is not None else None
        return get_runs_feed_entries(
            graphene_info=graphene_info, cursor=cursor, limit=limit, filters=selector, view=view
        )

    def resolve_runsFeedCountOrError(
        self,
        graphene_info: ResolveInfo,
        view: GrapheneRunsFeedView,
        filter: Optional[GrapheneRunsFilter] = None,  # noqa: A002
    ):
        selector = filter.to_selector() if filter is not None else None
        return GrapheneRunsFeedCount(
            get_runs_feed_count(
                graphene_info,
                selector,
                view=view,
            )
        )

    @capture_error
    def resolve_partitionSetsOrError(
        self,
        graphene_info: ResolveInfo,
        repositorySelector: RepositorySelector,
        pipelineName: str,
    ):
        return get_partition_sets_or_error(
            graphene_info,
            RepositorySelector.from_graphql_input(repositorySelector),
            pipelineName,
        )

    @capture_error
    def resolve_partitionSetOrError(
        self,
        graphene_info: ResolveInfo,
        repositorySelector: RepositorySelector,
        partitionSetName: Optional[str] = None,
    ):
        return get_partition_set(
            graphene_info,
            RepositorySelector.from_graphql_input(repositorySelector),
            # partitionSetName should prob be required
            partitionSetName,  # type: ignore
        )

    @capture_error
    def resolve_runTagKeysOrError(self, graphene_info: ResolveInfo):
        return get_run_tag_keys(graphene_info)

    @capture_error
    def resolve_runTagsOrError(
        self,
        graphene_info: ResolveInfo,
        tagKeys: list[str],
        valuePrefix: Optional[str] = None,
        limit: Optional[int] = None,
    ):
        return get_run_tags(graphene_info, tagKeys, valuePrefix, limit)

    @capture_error
    def resolve_runGroupOrError(self, graphene_info: ResolveInfo, runId):
        return get_run_group(graphene_info, runId)

    @capture_error
    def resolve_isPipelineConfigValid(
        self,
        graphene_info: ResolveInfo,
        pipeline: GraphenePipelineSelector,
        mode: str,
        runConfigData: Optional[Any] = None,  # custom scalar (GrapheneRunConfigData)
    ):
        return validate_pipeline_config(
            graphene_info,
            pipeline_selector_from_graphql(pipeline),
            parse_run_config_input(runConfigData or {}, raise_on_error=False),
        )

    @capture_error
    def resolve_executionPlanOrError(
        self,
        graphene_info: ResolveInfo,
        pipeline: GraphenePipelineSelector,
        mode: str,
        runConfigData: Optional[Any] = None,  # custom scalar (GrapheneRunConfigData)
    ):
        return get_execution_plan(
            graphene_info,
            pipeline_selector_from_graphql(pipeline),
            parse_run_config_input(runConfigData or {}, raise_on_error=True),  # type: ignore  # (possible str)
        )

    @capture_error
    def resolve_runConfigSchemaOrError(
        self,
        graphene_info: ResolveInfo,
        selector: GraphenePipelineSelector,
        mode: Optional[str] = None,
    ):
        return resolve_run_config_schema_or_error(
            graphene_info, pipeline_selector_from_graphql(selector), mode
        )

    def resolve_instance(self, graphene_info: ResolveInfo):
        return GrapheneInstance(graphene_info.context.instance)

    def resolve_assetNodes(
        self,
        graphene_info: ResolveInfo,
        loadMaterializations: bool,
        group: Optional[GrapheneAssetGroupSelector] = None,
        pipeline: Optional[GraphenePipelineSelector] = None,
        assetKeys: Optional[Sequence[GrapheneAssetKeyInput]] = None,
    ) -> Sequence[GrapheneAssetNode]:
        if assetKeys == []:
            return []
        elif not assetKeys:
            use_all_asset_keys = True
            resolved_asset_keys = None
        else:
            use_all_asset_keys = False
            resolved_asset_keys = set(
                AssetKey.from_graphql_input(asset_key_input) for asset_key_input in assetKeys or []
            )

        repo = None

        dynamic_partitions_loader = CachingDynamicPartitionsLoader(graphene_info.context.instance)
        if group is not None:
            group_name = group.groupName
            repo_sel = RepositorySelector.from_graphql_input(group)
            repo_loc_entry = graphene_info.context.get_location_entry(repo_sel.location_name)
            repo_loc = repo_loc_entry.code_location if repo_loc_entry else None
            if not repo_loc or not repo_loc.has_repository(repo_sel.repository_name):
                return []

            repo = repo_loc.get_repository(repo_sel.repository_name)
            remote_nodes = [
                remote_node
                for remote_node in repo.asset_graph.asset_nodes
                if remote_node.group_name == group_name
            ]
        elif pipeline is not None:
            selector = pipeline_selector_from_graphql(pipeline)
            remote_nodes = graphene_info.context.get_assets_in_job(selector)
        else:
            if not use_all_asset_keys and resolved_asset_keys:
                remote_nodes = [
                    graphene_info.context.asset_graph.get(asset_key)
                    for asset_key in resolved_asset_keys
                    if graphene_info.context.asset_graph.has(asset_key)
                ]
            else:
                remote_nodes = [
                    remote_node for remote_node in graphene_info.context.asset_graph.asset_nodes
                ]

        # Filter down to requested asset keys
        results = [
            remote_node
            for remote_node in remote_nodes
            if use_all_asset_keys or remote_node.key in check.not_none(resolved_asset_keys)
        ]

        if not results:
            return []

        final_keys = [node.key for node in results]
        AssetRecord.prepare(graphene_info.context, final_keys)

        def load_asset_graph() -> RemoteAssetGraph:
            if repo is not None:
                return repo.asset_graph
            else:
                return graphene_info.context.asset_graph

        stale_status_loader = StaleStatusLoader(
            instance=graphene_info.context.instance,
            asset_graph=load_asset_graph,
            loading_context=graphene_info.context,
        )

        nodes = [
            GrapheneAssetNode(
                remote_node=remote_node,
                stale_status_loader=stale_status_loader,
                dynamic_partitions_loader=dynamic_partitions_loader,
            )
            for remote_node in results
        ]
        return sorted(nodes, key=lambda node: node.id)

    def resolve_assetNodeOrError(self, graphene_info: ResolveInfo, assetKey: GrapheneAssetKeyInput):
        asset_key_input = cast("Mapping[str, Sequence[str]]", assetKey)
        return get_asset_node(graphene_info, AssetKey.from_graphql_input(asset_key_input))

    @capture_error
    def resolve_assetsOrError(
        self,
        graphene_info: ResolveInfo,
        prefix: Optional[Sequence[str]] = None,
        cursor: Optional[str] = None,
        limit: Optional[int] = None,
    ):
        return get_assets(
            graphene_info,
            prefix=prefix,
            cursor=cursor,
            limit=limit,
        )

    @capture_error
    def resolve_assetRecordsOrError(
        self,
        graphene_info: ResolveInfo,
        prefix: Optional[Sequence[str]] = None,
        cursor: Optional[str] = None,
        limit: Optional[int] = None,
    ):
        return get_asset_records(
            graphene_info,
            prefix=prefix,
            cursor=cursor,
            limit=limit,
        )

    def resolve_assetOrError(self, graphene_info: ResolveInfo, assetKey: GrapheneAssetKeyInput):
        return get_asset(graphene_info, AssetKey.from_graphql_input(assetKey))

    def resolve_assetNodeAdditionalRequiredKeys(
        self,
        graphene_info: ResolveInfo,
        assetKeys: Sequence[GrapheneAssetKeyInput],
    ):
        assert assetKeys is not None
        raw_asset_keys = cast("Sequence[Mapping[str, Sequence[str]]]", assetKeys)
        asset_keys = set(AssetKey.from_graphql_input(asset_key) for asset_key in raw_asset_keys)
        return get_additional_required_keys(graphene_info, asset_keys)

    def resolve_assetNodeDefinitionCollisions(
        self,
        graphene_info: ResolveInfo,
        assetKeys: Sequence[GrapheneAssetKeyInput],
    ):
        assert assetKeys is not None
        raw_asset_keys = cast("Sequence[Mapping[str, Sequence[str]]]", assetKeys)
        asset_keys = set(AssetKey.from_graphql_input(asset_key) for asset_key in raw_asset_keys)
        return get_asset_node_definition_collisions(graphene_info, asset_keys)

    @capture_error
    def resolve_partitionBackfillOrError(self, graphene_info: ResolveInfo, backfillId: str):
        return get_backfill(graphene_info, backfillId)

    @capture_error
    def resolve_partitionBackfillsOrError(
        self,
        graphene_info: ResolveInfo,
        status: Optional[GrapheneBulkActionStatus] = None,
        cursor: Optional[str] = None,
        limit: Optional[int] = None,
        filters: Optional[GrapheneBulkActionsFilter] = None,
    ):
        return get_backfills(
            graphene_info,
            status=BulkActionStatus.from_graphql_input(status) if status else None,
            cursor=cursor,
            limit=limit,
            filters=filters.to_selector() if filters else None,
        )

    def resolve_assetBackfillPreview(
        self, graphene_info: ResolveInfo, params: GrapheneAssetBackfillPreviewParams
    ) -> Sequence[GrapheneAssetPartitions]:
        return get_asset_backfill_preview(graphene_info, params)

    def resolve_permissions(self, graphene_info: ResolveInfo):
        permissions = graphene_info.context.permissions
        return [GraphenePermission(permission, value) for permission, value in permissions.items()]

    def resolve_canBulkTerminate(self, graphene_info: ResolveInfo) -> bool:
        return graphene_info.context.has_permission(Permissions.TERMINATE_PIPELINE_EXECUTION)

    def resolve_assetsLatestInfo(
        self,
        graphene_info: ResolveInfo,
        assetKeys: Sequence[GrapheneAssetKeyInput],
    ):
        asset_keys = set(AssetKey.from_graphql_input(asset_key) for asset_key in assetKeys)

        remote_nodes = {
            graphene_info.context.asset_graph.get(asset_key)
            for asset_key in asset_keys
            if graphene_info.context.asset_graph.has(asset_key)
        }

        # Build mapping of asset key to the step keys required to generate the asset
        step_keys_by_asset: dict[AssetKey, Sequence[str]] = {
            remote_node.key: remote_node.resolve_to_singular_repo_scoped_node().asset_node_snap.op_names
            for remote_node in remote_nodes
        }

        AssetRecord.prepare(graphene_info.context, asset_keys)

        return get_assets_latest_info(graphene_info, step_keys_by_asset)

    @capture_error
    def resolve_logsForRun(
        self,
        graphene_info: ResolveInfo,
        runId: str,
        afterCursor: Optional[str] = None,
        limit: Optional[int] = None,
    ):
        return get_logs_for_run(graphene_info, runId, afterCursor, limit)

    def resolve_capturedLogsMetadata(
        self, graphene_info: ResolveInfo, logKey: Sequence[str]
    ) -> GrapheneCapturedLogsMetadata:
        return get_captured_log_metadata(graphene_info, logKey)

    def resolve_capturedLogs(
        self,
        graphene_info: ResolveInfo,
        logKey: Sequence[str],
        cursor: Optional[str] = None,
        limit: Optional[int] = None,
    ) -> GrapheneCapturedLogs:
        log_data = get_compute_log_manager(graphene_info).get_log_data(
            logKey, cursor=cursor, max_bytes=limit
        )
        return from_captured_log_data(log_data)

    def resolve_shouldShowNux(self, graphene_info):
        return graphene_info.context.instance.nux_enabled and not get_has_seen_nux()

    def resolve_test(self, _):
        return GrapheneTestFields()

    def resolve_autoMaterializeAssetEvaluationsOrError(
        self,
        graphene_info: ResolveInfo,
        assetKey: GrapheneAssetKeyInput,
        limit: int,
        cursor: Optional[str] = None,
    ):
        return fetch_auto_materialize_asset_evaluations(
            graphene_info=graphene_info,
            graphene_asset_key=assetKey,
            cursor=cursor,
            limit=limit,
        )

    def resolve_autoMaterializeEvaluationsForEvaluationId(
        self,
        graphene_info: ResolveInfo,
        evaluationId: int,
    ):
        return fetch_auto_materialize_asset_evaluations_for_evaluation_id(
            graphene_info=graphene_info, evaluation_id=evaluationId
        )

    def resolve_assetConditionEvaluationForPartition(
        self,
        graphene_info: ResolveInfo,
        assetKey: Optional[GrapheneAssetKeyInput],
        evaluationId: str,
        partition: str,
    ):
        return fetch_asset_condition_evaluation_record_for_partition(
            graphene_info=graphene_info,
            graphene_asset_key=assetKey,
            evaluation_id=int(evaluationId),
            partition_key=partition,
        )

    def resolve_assetConditionEvaluationRecordsOrError(
        self,
        graphene_info: ResolveInfo,
        assetKey: Optional[GrapheneAssetKeyInput],
        limit: int,
        cursor: Optional[str] = None,
        assetCheckKey: Optional[GrapheneAssetCheckHandleInput] = None,
    ):
        return fetch_asset_condition_evaluation_records_for_asset_key(
            graphene_info=graphene_info,
            graphene_entity_key=check.not_none(assetKey or assetCheckKey),
            cursor=cursor,
            limit=limit,
        )

    def resolve_truePartitionsForAutomationConditionEvaluationNode(
        self,
        graphene_info: ResolveInfo,
        assetKey: Optional[GrapheneAssetKeyInput],
        evaluationId: str,
        nodeUniqueId: str,
    ):
        return fetch_true_partitions_for_evaluation_node(
            graphene_info=graphene_info,
            graphene_entity_key=assetKey,
            evaluation_id=int(evaluationId),
            node_unique_id=nodeUniqueId,
        )

    def resolve_assetConditionEvaluationsForEvaluationId(
        self, graphene_info: ResolveInfo, evaluationId: int
    ):
        return fetch_asset_condition_evaluation_records_for_evaluation_id(
            graphene_info=graphene_info, evaluation_id=evaluationId
        )

    def resolve_autoMaterializeTicks(
        self,
        graphene_info,
        dayRange=None,
        dayOffset=None,
        limit=None,
        cursor=None,
        statuses=None,
        beforeTimestamp=None,
        afterTimestamp=None,
    ):
        # Only valid for ticks from before auto-materialize was moved to be powered by multiple
        # sensors
        from dagster._daemon.asset_daemon import (
            _PRE_SENSOR_AUTO_MATERIALIZE_ORIGIN_ID,
            _PRE_SENSOR_AUTO_MATERIALIZE_SELECTOR_ID,
        )

        return get_instigation_ticks(
            graphene_info=graphene_info,
            instigator_type=InstigatorType.AUTO_MATERIALIZE,
            instigator_origin_id=_PRE_SENSOR_AUTO_MATERIALIZE_ORIGIN_ID,
            selector_id=_PRE_SENSOR_AUTO_MATERIALIZE_SELECTOR_ID,
            batch_loader=None,
            dayRange=dayRange,
            dayOffset=dayOffset,
            limit=limit,
            cursor=cursor,
            status_strings=statuses,
            before=beforeTimestamp,
            after=afterTimestamp,
        )

    def resolve_assetCheckExecutions(
        self,
        graphene_info: ResolveInfo,
        assetKey: GrapheneAssetKeyInput,
        checkName: str,
        limit: int,
        cursor: Optional[str] = None,
    ):
        return fetch_asset_check_executions(
            graphene_info.context,
            asset_check_key=AssetCheckKey(
                asset_key=AssetKey.from_graphql_input(assetKey), name=checkName
            ),
            limit=limit,
            cursor=cursor,
        )



import time
from collections.abc import Sequence
from typing import TYPE_CHECKING, Optional

import dagster._check as check
from dagster._core.definitions.run_request import InstigatorType
from dagster._core.definitions.selector import (
    JobSubsetSelector,
    RepositorySelector,
    ScheduleSelector,
)
from dagster._core.scheduler.instigation import InstigatorState, InstigatorStatus
from dagster._core.workspace.permissions import Permissions

from dagster_graphql.implementation.loader import RepositoryScopedBatchLoader
from dagster_graphql.implementation.utils import (
    UserFacingGraphQLError,
    assert_permission,
    assert_permission_for_location,
)
from dagster_graphql.schema.util import ResolveInfo

if TYPE_CHECKING:
    from dagster_graphql.schema.instigation import GrapheneDryRunInstigationTick
    from dagster_graphql.schema.schedules import (
        GrapheneSchedule,
        GrapheneScheduler,
        GrapheneSchedules,
        GrapheneScheduleStateResult,
    )


def start_schedule(
    graphene_info: ResolveInfo, schedule_selector: ScheduleSelector
) -> "GrapheneScheduleStateResult":
    from dagster_graphql.schema.errors import GrapheneScheduleNotFoundError
    from dagster_graphql.schema.instigation import GrapheneInstigationState
    from dagster_graphql.schema.schedules import GrapheneScheduleStateResult

    check.inst_param(schedule_selector, "schedule_selector", ScheduleSelector)
    schedule = graphene_info.context.get_schedule(schedule_selector)
    if not schedule:
        raise UserFacingGraphQLError(
            GrapheneScheduleNotFoundError(schedule_name=schedule_selector.schedule_name)
        )

    stored_state = graphene_info.context.instance.start_schedule(schedule)
    schedule_state = schedule.get_current_instigator_state(stored_state)

    return GrapheneScheduleStateResult(GrapheneInstigationState(schedule_state))


def stop_schedule(
    graphene_info: ResolveInfo, schedule_origin_id: str, schedule_selector_id: str
) -> "GrapheneScheduleStateResult":
    from dagster_graphql.schema.instigation import GrapheneInstigationState
    from dagster_graphql.schema.schedules import GrapheneScheduleStateResult

    instance = graphene_info.context.instance

    schedules = {
        schedule.get_remote_origin_id(): schedule
        for repository_location in graphene_info.context.code_locations
        for repository in repository_location.get_repositories().values()
        for schedule in repository.get_schedules()
    }

    schedule = schedules.get(schedule_origin_id)

    if schedule:
        assert_permission_for_location(
            graphene_info,
            Permissions.STOP_RUNNING_SCHEDULE,
            schedule.selector.location_name,
        )
    else:
        assert_permission(
            graphene_info,
            Permissions.STOP_RUNNING_SCHEDULE,
        )

    schedule_state = instance.stop_schedule(schedule_origin_id, schedule_selector_id, schedule)
    return GrapheneScheduleStateResult(GrapheneInstigationState(schedule_state))


def reset_schedule(
    graphene_info: ResolveInfo, schedule_selector: ScheduleSelector
) -> "GrapheneScheduleStateResult":
    from dagster_graphql.schema.instigation import GrapheneInstigationState
    from dagster_graphql.schema.schedules import GrapheneScheduleStateResult

    check.inst_param(schedule_selector, "schedule_selector", ScheduleSelector)

    location = graphene_info.context.get_code_location(schedule_selector.location_name)
    repository = location.get_repository(schedule_selector.repository_name)

    schedule = repository.get_schedule(schedule_selector.schedule_name)
    stored_state = graphene_info.context.instance.reset_schedule(schedule)
    schedule_state = schedule.get_current_instigator_state(stored_state)

    return GrapheneScheduleStateResult(GrapheneInstigationState(schedule_state))


def get_scheduler_or_error(graphene_info: ResolveInfo) -> "GrapheneScheduler":
    from dagster_graphql.schema.errors import GrapheneSchedulerNotDefinedError
    from dagster_graphql.schema.schedules import GrapheneScheduler

    instance = graphene_info.context.instance

    if not instance.scheduler:
        raise UserFacingGraphQLError(GrapheneSchedulerNotDefinedError())

    return GrapheneScheduler(scheduler_class=instance.scheduler.__class__.__name__)


def get_schedules_or_error(
    graphene_info: ResolveInfo,
    repository_selector: RepositorySelector,
    instigator_statuses: Optional[set[InstigatorStatus]] = None,
) -> "GrapheneSchedules":
    from dagster_graphql.schema.schedules import GrapheneSchedule, GrapheneSchedules

    check.inst_param(repository_selector, "repository_selector", RepositorySelector)

    location = graphene_info.context.get_code_location(repository_selector.location_name)
    repository = location.get_repository(repository_selector.repository_name)
    batch_loader = RepositoryScopedBatchLoader(graphene_info.context.instance, repository)
    schedules = repository.get_schedules()
    schedule_states = graphene_info.context.instance.all_instigator_state(
        repository_origin_id=repository.get_remote_origin_id(),
        repository_selector_id=repository_selector.selector_id,
        instigator_type=InstigatorType.SCHEDULE,
        instigator_statuses=instigator_statuses,
    )

    schedule_states_by_name = {state.name: state for state in schedule_states}
    if instigator_statuses:
        filtered = [
            schedule
            for schedule in schedules
            if schedule.get_current_instigator_state(
                schedule_states_by_name.get(schedule.name)
            ).status
            in instigator_statuses
        ]
    else:
        filtered = schedules

    results = [
        GrapheneSchedule(
            schedule,
            schedule_states_by_name.get(schedule.name),
            batch_loader,
        )
        for schedule in filtered
    ]

    return GrapheneSchedules(results=results)


def get_schedules_for_job(
    graphene_info: ResolveInfo, selector: JobSubsetSelector
) -> Sequence["GrapheneSchedule"]:
    from dagster_graphql.schema.schedules import GrapheneSchedule

    check.inst_param(selector, "selector", JobSubsetSelector)

    schedules = graphene_info.context.get_schedules_targeting_job(selector)

    results = []
    for schedule in schedules:
        schedule_state = graphene_info.context.instance.get_instigator_state(
            schedule.get_remote_origin_id(),
            schedule.selector_id,
        )

        results.append(GrapheneSchedule(schedule, schedule_state))

    return results


def get_schedule_or_error(
    graphene_info: ResolveInfo, schedule_selector: ScheduleSelector
) -> "GrapheneSchedule":
    from dagster_graphql.schema.errors import GrapheneScheduleNotFoundError
    from dagster_graphql.schema.schedules import GrapheneSchedule

    check.inst_param(schedule_selector, "schedule_selector", ScheduleSelector)

    schedule = graphene_info.context.get_schedule(schedule_selector)
    if not schedule:
        raise UserFacingGraphQLError(
            GrapheneScheduleNotFoundError(schedule_name=schedule_selector.schedule_name)
        )

    schedule_state = graphene_info.context.instance.get_instigator_state(
        schedule.get_remote_origin_id(), schedule.selector_id
    )
    return GrapheneSchedule(schedule, schedule_state)


def get_schedule_next_tick(
    graphene_info: ResolveInfo, schedule_state: InstigatorState
) -> Optional["GrapheneDryRunInstigationTick"]:
    from dagster_graphql.schema.instigation import GrapheneDryRunInstigationTick

    if not schedule_state.is_running:
        return None

    repository_origin = schedule_state.origin.repository_origin
    selector = ScheduleSelector(
        location_name=repository_origin.code_location_origin.location_name,
        repository_name=repository_origin.repository_name,
        schedule_name=schedule_state.name,
    )

    schedule = graphene_info.context.get_schedule(selector)
    if not schedule:
        return None

    time_iter = schedule.execution_time_iterator(time.time())

    next_timestamp = next(time_iter).timestamp()
    return GrapheneDryRunInstigationTick(schedule.schedule_selector, next_timestamp)



import asyncio
from typing import TYPE_CHECKING, Optional, cast

import graphene
from dagster import _check as check
from dagster._core.definitions.partition import CachingDynamicPartitionsLoader
from dagster._core.definitions.sensor_definition import SensorType
from dagster._core.remote_representation import (
    CodeLocation,
    GrpcServerCodeLocation,
    RemoteRepository,
)
from dagster._core.remote_representation.feature_flags import get_feature_flags_for_location
from dagster._core.remote_representation.grpc_server_state_subscriber import (
    LocationStateChangeEvent,
    LocationStateChangeEventType,
    LocationStateSubscriber,
)
from dagster._core.remote_representation.handle import RepositoryHandle
from dagster._core.workspace.context import WorkspaceProcessContext
from dagster._core.workspace.workspace import CodeLocationEntry, CodeLocationLoadStatus
from dagster.components.core.load_defs import PLUGIN_COMPONENT_TYPES_JSON_METADATA_KEY

from dagster_graphql.implementation.fetch_solids import get_solid, get_solids
from dagster_graphql.implementation.loader import RepositoryScopedBatchLoader, StaleStatusLoader
from dagster_graphql.implementation.utils import capture_error
from dagster_graphql.schema.asset_graph import GrapheneAssetGroup, GrapheneAssetNode
from dagster_graphql.schema.env_vars import (
    GrapheneLocationDocsJson,
    GrapheneLocationDocsJsonOrError,
)
from dagster_graphql.schema.errors import GraphenePythonError, GrapheneRepositoryNotFoundError
from dagster_graphql.schema.partition_sets import GraphenePartitionSet
from dagster_graphql.schema.permissions import GraphenePermission
from dagster_graphql.schema.pipelines.pipeline import GrapheneJob, GraphenePipeline
from dagster_graphql.schema.repository_origin import (
    GrapheneRepositoryMetadata,
    GrapheneRepositoryOrigin,
)
from dagster_graphql.schema.resources import GrapheneResourceDetails
from dagster_graphql.schema.schedules import GrapheneSchedule
from dagster_graphql.schema.sensors import GrapheneSensor, GrapheneSensorType
from dagster_graphql.schema.used_solid import GrapheneUsedSolid
from dagster_graphql.schema.util import ResolveInfo, non_null_list

if TYPE_CHECKING:
    from dagster._core.remote_representation.external_data import AssetNodeSnap

GrapheneLocationStateChangeEventType = graphene.Enum.from_enum(LocationStateChangeEventType)


class GrapheneDagsterLibraryVersion(graphene.ObjectType):
    name = graphene.NonNull(graphene.String)
    version = graphene.NonNull(graphene.String)

    class Meta:
        name = "DagsterLibraryVersion"


class GrapheneRepositoryLocationLoadStatus(graphene.Enum):
    LOADING = "LOADING"
    LOADED = "LOADED"

    class Meta:
        name = "RepositoryLocationLoadStatus"

    @classmethod
    def from_python_status(cls, python_status):
        check.inst_param(python_status, "python_status", CodeLocationLoadStatus)
        if python_status == CodeLocationLoadStatus.LOADING:
            return GrapheneRepositoryLocationLoadStatus.LOADING
        elif python_status == CodeLocationLoadStatus.LOADED:
            return GrapheneRepositoryLocationLoadStatus.LOADED
        else:
            check.failed(f"Invalid location load status: {python_status}")


class GrapheneRepositoryLocation(graphene.ObjectType):
    id = graphene.NonNull(graphene.ID)
    name = graphene.NonNull(graphene.String)
    is_reload_supported = graphene.NonNull(graphene.Boolean)
    environment_path = graphene.String()
    repositories = non_null_list(lambda: GrapheneRepository)
    server_id = graphene.String()
    dagsterLibraryVersions = graphene.List(graphene.NonNull(GrapheneDagsterLibraryVersion))

    class Meta:
        name = "RepositoryLocation"

    def __init__(self, name: str, location: Optional[CodeLocation] = None):
        self._location = location
        super().__init__(
            name=name,
        )

    def resolve_id(self, _) -> str:
        return self.name

    def get_location(self, graphene_info: ResolveInfo) -> CodeLocation:
        if self._location is None:
            self._location = graphene_info.context.get_code_location(self.name)
        return self._location

    def resolve_repositories(self, graphene_info: ResolveInfo):
        return [
            GrapheneRepository(repository.handle)
            for repository in self.get_location(graphene_info).get_repositories().values()
        ]

    def resolve_dagsterLibraryVersions(self, graphene_info: ResolveInfo):
        libs = self.get_location(graphene_info).get_dagster_library_versions()
        if libs is None:
            return None

        return [GrapheneDagsterLibraryVersion(name, ver) for name, ver in libs.items()]

    def resolve_server_id(self, graphene_info: ResolveInfo):
        location = self.get_location(graphene_info)
        return location.server_id if isinstance(location, GrpcServerCodeLocation) else None

    def resolve_is_reload_supported(self, graphene_info: ResolveInfo):
        location = self.get_location(graphene_info)
        return location.is_reload_supported


class GrapheneRepositoryLocationOrLoadError(graphene.Union):
    class Meta:
        types = (
            GrapheneRepositoryLocation,
            GraphenePythonError,
        )
        name = "RepositoryLocationOrLoadError"


class GrapheneWorkspaceLocationStatusEntry(graphene.ObjectType):
    id = graphene.NonNull(graphene.ID)
    name = graphene.NonNull(graphene.String)
    loadStatus = graphene.NonNull(GrapheneRepositoryLocationLoadStatus)
    updateTimestamp = graphene.NonNull(graphene.Float)
    versionKey = graphene.NonNull(graphene.String)

    permissions = graphene.Field(non_null_list(GraphenePermission))

    class Meta:
        name = "WorkspaceLocationStatusEntry"

    def __init__(
        self,
        id,
        name,
        load_status,
        update_timestamp,
        version_key,
    ):
        super().__init__(
            id=id,
            name=name,
            loadStatus=load_status,
            updateTimestamp=update_timestamp,
            versionKey=version_key,
        )

    def resolve_permissions(self, graphene_info):
        permissions = graphene_info.context.permissions_for_location(location_name=self.name)
        return [GraphenePermission(permission, value) for permission, value in permissions.items()]


class GrapheneWorkspaceLocationStatusEntries(graphene.ObjectType):
    entries = non_null_list(GrapheneWorkspaceLocationStatusEntry)

    class Meta:
        name = "WorkspaceLocationStatusEntries"


class GrapheneWorkspaceLocationStatusEntriesOrError(graphene.Union):
    class Meta:
        types = (
            GrapheneWorkspaceLocationStatusEntries,
            GraphenePythonError,
        )
        name = "WorkspaceLocationStatusEntriesOrError"


class GrapheneFeatureFlag(graphene.ObjectType):
    class Meta:
        name = "FeatureFlag"

    name = graphene.NonNull(graphene.String)
    enabled = graphene.NonNull(graphene.Boolean)


class GrapheneWorkspaceLocationEntry(graphene.ObjectType):
    id = graphene.NonNull(graphene.ID)
    name = graphene.NonNull(graphene.String)
    locationOrLoadError = graphene.Field(GrapheneRepositoryLocationOrLoadError)
    loadStatus = graphene.NonNull(GrapheneRepositoryLocationLoadStatus)
    displayMetadata = non_null_list(GrapheneRepositoryMetadata)
    updatedTimestamp = graphene.NonNull(graphene.Float)
    versionKey = graphene.NonNull(graphene.String)

    permissions = graphene.Field(non_null_list(GraphenePermission))

    featureFlags = non_null_list(GrapheneFeatureFlag)

    class Meta:
        name = "WorkspaceLocationEntry"

    def __init__(self, location_entry: CodeLocationEntry):
        self._location_entry = check.inst_param(location_entry, "location_entry", CodeLocationEntry)
        super().__init__(name=self._location_entry.origin.location_name)

    def resolve_id(self, _):
        return self.name

    def resolve_locationOrLoadError(self, _: ResolveInfo):
        if self._location_entry.code_location:
            return GrapheneRepositoryLocation(
                self._location_entry.code_location.name,
                self._location_entry.code_location,
            )

        error = self._location_entry.load_error
        return GraphenePythonError(error) if error else None

    def resolve_loadStatus(self, _):
        return GrapheneRepositoryLocationLoadStatus.from_python_status(
            self._location_entry.load_status
        )

    def resolve_displayMetadata(self, _):
        metadata = self._location_entry.display_metadata
        return [
            GrapheneRepositoryMetadata(key=key, value=value)
            for key, value in metadata.items()
            if value is not None
        ]

    def resolve_updatedTimestamp(self, _) -> float:
        return self._location_entry.update_timestamp

    def resolve_versionKey(self, _) -> str:
        return self._location_entry.version_key

    def resolve_permissions(self, graphene_info):
        permissions = graphene_info.context.permissions_for_location(location_name=self.name)
        return [GraphenePermission(permission, value) for permission, value in permissions.items()]

    def resolve_featureFlags(self, graphene_info):
        feature_flags = get_feature_flags_for_location(self._location_entry)
        return [
            GrapheneFeatureFlag(name=feature_flag_name.value, enabled=feature_flag_enabled)
            for feature_flag_name, feature_flag_enabled in feature_flags.items()
        ]


class GrapheneRepository(graphene.ObjectType):
    id = graphene.NonNull(graphene.ID)
    name = graphene.NonNull(graphene.String)
    location = graphene.NonNull(GrapheneRepositoryLocation)
    pipelines = non_null_list(GraphenePipeline)
    jobs = non_null_list(GrapheneJob)
    usedSolids = graphene.Field(non_null_list(GrapheneUsedSolid))
    usedSolid = graphene.Field(GrapheneUsedSolid, name=graphene.NonNull(graphene.String))
    origin = graphene.NonNull(GrapheneRepositoryOrigin)
    partitionSets = non_null_list(GraphenePartitionSet)
    schedules = non_null_list(GrapheneSchedule)
    sensors = graphene.Field(
        non_null_list(GrapheneSensor), sensorType=graphene.Argument(GrapheneSensorType)
    )
    assetNodes = non_null_list(GrapheneAssetNode)
    displayMetadata = non_null_list(GrapheneRepositoryMetadata)
    assetGroups = non_null_list(GrapheneAssetGroup)
    allTopLevelResourceDetails = non_null_list(GrapheneResourceDetails)
    hasLocationDocs = graphene.Field(
        graphene.NonNull(graphene.Boolean),
        description="Retrieves whether the code location has integrated docs.",
    )

    locationDocsJsonOrError = graphene.Field(
        graphene.NonNull(GrapheneLocationDocsJsonOrError),
        description="Retrieves JSON blob to drive integrated code location docs.",
    )

    class Meta:
        name = "Repository"

    def __init__(
        self,
        handle: RepositoryHandle,
    ):
        # Warning! GrapheneAssetNode contains a GrapheneRepository. Any computation in this
        # __init__ will be done **once per asset**. Ensure that any expensive work is done
        # elsewhere or cached.
        self._handle = handle

        self._batch_loader = None

        super().__init__(name=handle.repository_name)

    def get_repository(self, graphene_info: ResolveInfo) -> RemoteRepository:
        return graphene_info.context.get_repository(self._handle.to_selector())

    def get_batch_loader(self, graphene_info: ResolveInfo):
        if self._batch_loader is None:
            self._batch_loader = RepositoryScopedBatchLoader(
                graphene_info.context.instance, self.get_repository(graphene_info)
            )
        return self._batch_loader

    def resolve_id(self, _graphene_info: ResolveInfo) -> str:
        return self._handle.get_compound_id().to_string()

    def resolve_origin(self, _graphene_info: ResolveInfo):
        origin = self._handle.get_remote_origin()
        return GrapheneRepositoryOrigin(origin)

    def resolve_location(self, graphene_info: ResolveInfo):
        return GrapheneRepositoryLocation(self._handle.location_name)

    def resolve_schedules(self, graphene_info: ResolveInfo):
        batch_loader = self.get_batch_loader(graphene_info)
        repository = self.get_repository(graphene_info)
        return sorted(
            [
                GrapheneSchedule(
                    schedule,
                    batch_loader.get_schedule_state(schedule.name),
                    batch_loader,
                )
                for schedule in repository.get_schedules()
            ],
            key=lambda schedule: schedule.name,
        )

    def resolve_sensors(self, graphene_info: ResolveInfo, sensorType: Optional[SensorType] = None):
        batch_loader = self.get_batch_loader(graphene_info)
        repository = self.get_repository(graphene_info)
        return [
            GrapheneSensor(
                sensor,
                batch_loader.get_sensor_state(sensor.name),
                batch_loader,
            )
            for sensor in sorted(repository.get_sensors(), key=lambda sensor: sensor.name)
            if not sensorType or sensor.sensor_type == sensorType
        ]

    def resolve_pipelines(self, graphene_info: ResolveInfo):
        return [
            GraphenePipeline(pipeline)
            for pipeline in sorted(
                self.get_repository(graphene_info).get_all_jobs(),
                key=lambda pipeline: pipeline.name,
            )
        ]

    def resolve_jobs(self, graphene_info: ResolveInfo):
        return [
            GrapheneJob(pipeline)
            for pipeline in sorted(
                self.get_repository(graphene_info).get_all_jobs(),
                key=lambda pipeline: pipeline.name,
            )
        ]

    def resolve_usedSolid(self, graphene_info: ResolveInfo, name):
        return get_solid(self.get_repository(graphene_info), name)

    def resolve_usedSolids(self, graphene_info: ResolveInfo):
        return get_solids(self.get_repository(graphene_info))

    def resolve_partitionSets(self, graphene_info: ResolveInfo):
        return (
            GraphenePartitionSet(self._handle, partition_set)
            for partition_set in self.get_repository(graphene_info).get_partition_sets()
        )

    def resolve_displayMetadata(self, graphene_info: ResolveInfo):
        metadata = self._handle.display_metadata
        return [
            GrapheneRepositoryMetadata(key=key, value=value)
            for key, value in metadata.items()
            if value is not None
        ]

    def resolve_assetNodes(self, graphene_info: ResolveInfo):
        remote_nodes = self.get_repository(graphene_info).asset_graph.asset_nodes

        dynamic_partitions_loader = CachingDynamicPartitionsLoader(
            graphene_info.context.instance,
        )
        stale_status_loader = StaleStatusLoader(
            instance=graphene_info.context.instance,
            asset_graph=lambda: self.get_repository(graphene_info).asset_graph,
            loading_context=graphene_info.context,
        )

        return [
            GrapheneAssetNode(
                remote_node=remote_node,
                stale_status_loader=stale_status_loader,
                dynamic_partitions_loader=dynamic_partitions_loader,
            )
            for remote_node in remote_nodes
        ]

    def resolve_assetGroups(self, graphene_info: ResolveInfo):
        groups: dict[str, list[AssetNodeSnap]] = {}
        for asset_node_snap in self.get_repository(graphene_info).get_asset_node_snaps():
            if not asset_node_snap.group_name:
                continue
            asset_node_snaps = groups.setdefault(asset_node_snap.group_name, [])
            asset_node_snaps.append(asset_node_snap)

        return [
            GrapheneAssetGroup(
                f"{self._handle.location_name}-{self._handle.repository_name}-{group_name}",
                group_name,
                [external_node.asset_key for external_node in asset_node_snaps],
            )
            for group_name, asset_node_snaps in groups.items()
        ]

    def resolve_allTopLevelResourceDetails(self, graphene_info) -> list[GrapheneResourceDetails]:
        return [
            GrapheneResourceDetails(
                location_name=self._handle.location_name,
                repository_name=self._handle.repository_name,
                remote_resource=resource,
            )
            for resource in sorted(
                self.get_repository(graphene_info).get_resources(),
                key=lambda resource: resource.name,
            )
            if resource.is_top_level
        ]

    def resolve_hasLocationDocs(self, graphene_info: ResolveInfo):
        repository = self.get_repository(graphene_info)

        return bool(
            repository.repository_snap.metadata
            and repository.repository_snap.metadata.get(PLUGIN_COMPONENT_TYPES_JSON_METADATA_KEY)
        )

    @capture_error
    def resolve_locationDocsJsonOrError(
        self,
        graphene_info: ResolveInfo,
    ) -> GrapheneLocationDocsJson:
        repository = self.get_repository(graphene_info)
        plugin_docs_json = (
            cast(
                "list",
                repository.repository_snap.metadata.get(
                    PLUGIN_COMPONENT_TYPES_JSON_METADATA_KEY, [[]]
                ),
            )[0]
            if repository.repository_snap.metadata
            else []
        )

        return GrapheneLocationDocsJson(json=plugin_docs_json)


class GrapheneRepositoryConnection(graphene.ObjectType):
    nodes = non_null_list(GrapheneRepository)

    class Meta:
        name = "RepositoryConnection"


class GrapheneWorkspace(graphene.ObjectType):
    id = graphene.NonNull(graphene.String)
    locationEntries = non_null_list(GrapheneWorkspaceLocationEntry)

    class Meta:
        name = "Workspace"

    def resolve_id(self, _graphene_info: ResolveInfo):
        return "Workspace"


class GrapheneLocationStateChangeEvent(graphene.ObjectType):
    event_type = graphene.NonNull(GrapheneLocationStateChangeEventType)
    message = graphene.NonNull(graphene.String)
    location_name = graphene.NonNull(graphene.String)
    server_id = graphene.Field(graphene.String)

    class Meta:
        name = "LocationStateChangeEvent"


class GrapheneLocationStateChangeSubscription(graphene.ObjectType):
    event = graphene.Field(graphene.NonNull(GrapheneLocationStateChangeEvent))

    class Meta:
        name = "LocationStateChangeSubscription"


async def gen_location_state_changes(graphene_info: ResolveInfo):
    # This lives on the process context and is never modified/destroyed, so we can
    # access it directly
    context = graphene_info.context.process_context

    if not isinstance(context, WorkspaceProcessContext):
        return

    queue: asyncio.Queue[LocationStateChangeEvent] = asyncio.Queue()
    loop = asyncio.get_event_loop()

    def _enqueue(event):
        loop.call_soon_threadsafe(queue.put_nowait, event)

    token = context.add_state_subscriber(LocationStateSubscriber(_enqueue))
    try:
        while True:
            event = await queue.get()
            yield GrapheneLocationStateChangeSubscription(
                event=GrapheneLocationStateChangeEvent(
                    event_type=event.event_type,
                    location_name=event.location_name,
                    message=event.message,
                    server_id=event.server_id,
                ),
            )
    finally:
        context.rm_state_subscriber(token)


class GrapheneRepositoriesOrError(graphene.Union):
    class Meta:
        types = (
            GrapheneRepositoryConnection,
            GrapheneRepositoryNotFoundError,
            GraphenePythonError,
        )
        name = "RepositoriesOrError"


class GrapheneWorkspaceOrError(graphene.Union):
    class Meta:
        types = (GrapheneWorkspace, GraphenePythonError)
        name = "WorkspaceOrError"


class GrapheneRepositoryOrError(graphene.Union):
    class Meta:
        types = (
            GraphenePythonError,
            GrapheneRepository,
            GrapheneRepositoryNotFoundError,
        )
        name = "RepositoryOrError"


class GrapheneWorkspaceLocationEntryOrError(graphene.Union):
    class Meta:
        types = (GrapheneWorkspaceLocationEntry, GraphenePythonError)
        name = "WorkspaceLocationEntryOrError"


types = [
    GrapheneLocationStateChangeEvent,
    GrapheneLocationStateChangeEventType,
    GrapheneLocationStateChangeSubscription,
    GrapheneRepositoriesOrError,
    GrapheneRepository,
    GrapheneRepositoryConnection,
    GrapheneRepositoryLocation,
    GrapheneRepositoryOrError,
    GrapheneWorkspaceLocationEntryOrError,
]



from typing import cast

import graphene
from dagster._core.storage.compute_log_manager import ComputeLogManager
from dagster._core.workspace.context import WorkspaceRequestContext


class ResolveInfo(graphene.ResolveInfo):
    @property
    def context(self) -> WorkspaceRequestContext:
        return cast("WorkspaceRequestContext", super().context)


def non_null_list(of_type):
    return graphene.NonNull(graphene.List(graphene.NonNull(of_type)))


def get_compute_log_manager(graphene_info: ResolveInfo) -> ComputeLogManager:
    return graphene_info.context.instance.compute_log_manager



from collections.abc import Sequence
from typing import TYPE_CHECKING

from dagster_graphql.schema.util import ResolveInfo

if TYPE_CHECKING:
    from dagster_graphql.schema.logs.compute_logs import GrapheneCapturedLogsMetadata


def get_captured_log_metadata(
    graphene_info: ResolveInfo, log_key: Sequence[str]
) -> "GrapheneCapturedLogsMetadata":
    from dagster_graphql.schema.logs.compute_logs import GrapheneCapturedLogsMetadata

    metadata = graphene_info.context.instance.compute_log_manager.get_log_metadata(log_key)
    return GrapheneCapturedLogsMetadata(
        stdoutDownloadUrl=metadata.stdout_download_url,
        stdoutLocation=metadata.stdout_location,
        stderrDownloadUrl=metadata.stderr_download_url,
        stderrLocation=metadata.stderr_location,
    )
